#
# Copyright (C) 2020 Sylvain Marsat.
#
#

#Note some assumptions about relative code and data locations
import sys
import os
here=os.path.dirname(os.path.abspath(__file__))
sys.path.append(here+"/../..")

import json
import numpy as np

import lisabeta
import lisabeta.tools.pytools as pytools
import lisabeta.lisa.pyresponse as pyresponse
import lisabeta.lisa.pyLISAnoise as pyLISAnoise
import lisabeta.lisa.lisa as lisa


# Full list of physical params - infer_params are a subset
# TODO: generalize
listparams = [
    "Mchirp",
    "eta",
    "chi1",
    "chi2",
    "fstart",
    "dist",
    "inc",
    "phi",
    "lambda",
    "beta",
    "psi"]

# Default waveform params
waveform_params_default = {
    "minf": 1e-5,
    "maxf": 0.5,
    "fend": None,
    "tstart": 0.0,
    "tmin": 0.0,
    "tmax": 4.0,
    "phiref": 0.0,
    "fref_for_phiref": None,
    "toffset": 0.0,
    "TDI": "TDIAET",
    "acc": 1e-4,
    "order_fresnel_stencil": 0,
    "approximant": "IMRPhenomD",
    "LISAconst": "Proposal",
    "responseapprox": "full",
    "frozenLISA": False,
    "TDIrescaled": True,
    "LISAnoise": {
        "InstrumentalNoise": "SciRDv1",
        "WDbackground": False,
        "WDduration" : 0.0,
        "lowf_add_pm_noise_f0": 0.0,
        "lowf_add_pm_noise_alpha": 2.0
    }
}

# Default pymultinest params
run_params_musthave = ['outputfiles_basename']
run_params_default = {
    "zerolike": False,
    "n_live_points": 1000,
    "evidence_tolerance": 0.1,
    "sampling_efficiency": 0.5,
    "seed": -1,
    "resume": False,
    "max_iter": 0,
    "verbose": False,
    "write_output": True,
    "n_iter_before_update": 100,
    "importance_nested_sampling": False,
    "multimodal": False,
    "const_efficiency_mode": False
}

# Default parameter range for angular parameters
params_range_default = {
    "inc": [0., np.pi],
    "phi": [-np.pi, np.pi],
    "lambda": [-np.pi, np.pi],
    "beta": [-np.pi/2, np.pi/2],
    "psi": [0., np.pi]
}

# Utility to go from Mchirp-eta variables to m1-m2
def convert_params_m1m2(params_Mchirpeta):
    params_m1m2 = params_Mchirpeta.copy()
    Mchirp = params_m1m2.pop('Mchirp')
    eta = params_m1m2.pop('eta')
    m1 = pytools.m1ofMchirpeta(Mchirp, eta)
    m2 = pytools.m2ofMchirpeta(Mchirp, eta)
    params_m1m2['m1'] = m1
    params_m1m2['m2'] = m2
    return params_m1m2
# Utility to go from m1-m2 variables to Mchirp-eta
def convert_params_Mchirpeta(params_m1m2):
    params_Mchirpeta = params_m1m2.copy()
    m1 = params_Mchirpeta.pop('m1')
    m2 = params_Mchirpeta.pop('m2')
    Mchirp = pytools.Mchirpofm1m2(m1, m2)
    eta = pytools.etaofm1m2(m1, m2)
    params_Mchirpeta['Mchirp'] = Mchirp
    params_Mchirpeta['eta'] = eta
    return params_Mchirpeta

# Utilities for cube-to-param mapping
def map_cube_to_param(priortype, paramrange, x):
    if priortype=='uniform':
        return paramrange[0] + x * (paramrange[1] - paramrange[0])
    elif priortype=='sin': # NOTE: cos decreasing on [0,pi]
        cosrange = [np.cos(paramrange[1]), np.cos(paramrange[0])]
        return np.arccos(cosrange[0] + x * (cosrange[1] - cosrange[0]))
    elif priortype=='cos': # NOTE: sin increasing on [-pi/2,pi/2]
        sinrange = [np.sin(paramrange[0]), np.sin(paramrange[1])]
        return np.arcsin(sinrange[0] + x * (sinrange[1] - sinrange[0]))
    else:
        raise ValueError('priortype not recognized.')

def process_params(input_file_string):

    # Load json file with all parameters
    with open(input_file_string, 'r') as input_file:
        input_params = json.load(input_file)
    # Source params and prior params must be given
    source_params = input_params['source_params']
    prior_params = input_params['prior_params']
    # Waveform params and run params have predefined default values
    waveform_params = waveform_params_default.copy()
    waveform_params.update(input_params['waveform_params'])
    run_params = run_params_default.copy()
    run_params.update(input_params['run_params'])
    zerolike = run_params.pop('zerolike', False)
    # run params must contain output prefix for files -- no default here
    for musthave in run_params_musthave:
        if not (musthave in run_params):
            raise ValueError('run_params must contain %s.' % (musthave))

    # Use default param ranges to complete input ranges -- check completeness
    prior_type = prior_params['prior_type']
    params_range = prior_params['params_range']
    infer_params = prior_params['infer_params']
    for i, param in enumerate(infer_params):
        if params_range[i] == []:
            params_range[i] = params_range_default[param]
    if [] in params_range:
        raise ValueError('Some parameter ranges are neither given \
                      nor covered by the defaults.')

    # Convert in-place LISA constellation from string to dict
    LISAconst = pyresponse.LISAconstDict[waveform_params['LISAconst']]
    waveform_params['LISAconst'] = LISAconst
    # Convert in-place LISA instrumental noise from string to internal enum
    LISAnoise = waveform_params['LISAnoise']
    LISAnoise['InstrumentalNoise'] = pyLISAnoise.LISAnoiseDict[
                                                 LISAnoise['InstrumentalNoise']]

    # Convert source params to standard m1-m2 physical params
    # TODO: generalize
    source_params_Mchirpeta = convert_params_Mchirpeta(source_params)

    # Generate injection signal
    tdisignal_inj = lisa.GenerateLISATDISignal_SOBH(source_params,
                                                    **waveform_params)
    print("injection SNR=",tdisignal_inj['SNR'])

    # Parameters inferred and parameters fixed
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in listparams if p not in infer_params]
    n_dim = len(infer_params)
    n_fixed = len(fixed_params)
    # NOTE: should be able to set n_params and carry along extra fixed params
    # but did not succeed: was getting undefined behaviour in outputs
    n_params = n_dim

    # Wrapped parameters
    wrapped_params = [0] * n_dim
    for i in range(n_dim):
        wrapped_params[i] = int(prior_params['wrap_params'][i])
    run_params['wrapped_params'] = wrapped_params
    return n_params, prior_type, params_range, zerolike, infer_params, fixed_params, source_params_Mchirpeta, tdisignal_inj, waveform_params, run_params

if __name__ == '__main__':
    import pymultinest
    import argparse
    parser = argparse.ArgumentParser(description='Run SOBH parameter inference \
                                              using lisabeta and pymultinest.')

    parser.add_argument('input_file', type=str, help='Input json file \
                                                  with all parameters.')
    args = parser.parse_args()

    n_params, prior_type, params_range, zerolike, infer_params, fixed_params, source_params_Mchirpeta, tdisignal_inj, waveform_params, run_params = process_params(args.input_file)
    n_dim=n_params

    # Define function mapping unit cube to physical params according to priors
    # NOTE: this function is in-place for cube and does not return
    def cubeparams_to_physparams(cube, ndim, nparams):
        assert nparams==n_params and ndim==n_dim
        for i in range(n_dim):
            cube[i] = map_cube_to_param(prior_type[i], params_range[i], cube[i])

    # def test_cubeparams_to_physparams(cube, ndim, nparams):
    # 	assert nparams==n_params and ndim==n_params
    # 	cube_phys = [0] * n_params
    # 	for i in range(n_params):
    # 		cube_phys[i] = map_cube_to_param(prior_type[i], params_range[i], cube[i])
    # 	return cube_phys

    # Define loglikelihood function
    # NOTE: here cube contains already physical parameters, not unit cube values
    def loglike(cube, ndim, nparams):
        assert nparams==n_params and ndim==n_dim
        if zerolike: return 0.
        template_params = {}
        for i in range(n_dim):
            template_params[infer_params[i]] = cube[i]
        for i in range(len(fixed_params)):
            template_params[fixed_params[i]] = \
                                    source_params_Mchirpeta[fixed_params[i]]
        #print(template_params)
        template_params_m1m2 = convert_params_m1m2(template_params)
        lnL = lisa.LogLikelihoodLISA_SOBH(template_params_m1m2, tdisignal_inj,
                                          **waveform_params)
        #print(lnL)
        return lnL

    # A la grace de dieu
    pymultinest.run(loglike, cubeparams_to_physparams, n_dim, n_params=n_params,
                    **run_params)



#
# Copyright (C) 2020 Sylvain Marsat.
#
#


"""
    Python functions for ptemcee simulated inference of MBHBs with LISA.
"""


import os
import sys
import copy
import json
import h5py
import argparse
import numpy as np

import lisabeta
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools
import lisabeta.lisa.lisatools as lisatools
import lisabeta.lisa.snrtools as snrtools
import lisabeta.lisa.lisa as lisa
import lisabeta.lisa.lisa_fisher as lisa_fisher
import lisabeta.inference.inference as inference

import ptemcee
import ptemcee.mpi_pool as mpi_pool

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None


################################################################################
# Parse arguments, default params

# Parse arguments
parser = argparse.ArgumentParser(description='Run SMBH parameter inference \
                                              using lisabeta and ptemcee with sky jumps.')
parser.add_argument('input_file', type=str, help='Input json file \
                                                  with all parameters.')
args = parser.parse_args()

# Full list of physical params - infer_params are a subset
# TODO: generalize to other choices of mass and spin parameters
# TODO: at the moment, infer_params can only be a subset of this, not flexible
list_params_default = [
    "Mchirp",
    "q",
    "chi1",
    "chi2",
    "Deltat",
    "dist",
    "inc",
    "phi",
    "lambda",
    "beta",
    "psi"]

# Default waveform params
waveform_params_default = {
    "minf": 1e-5,
    "maxf": 0.5,
    "t0": 0.0,
    "timetomerger_max": 1.0,
    "fend": None,
    "tmin": None,
    "tmax": None,
    "Mfmax_model": 0.2,
    "phiref": 0.0,
    "fref_for_phiref": 0.0,
    "tref": 0.0,
    "fref_for_tref": 0.0,
    "force_phiref_fref": True,
    "toffset": 0.0,
    "modes": None,
    "TDI": "TDIAET",
    "acc": 1e-4,
    "order_fresnel_stencil": 0,
    "approximant": "IMRPhenomD",
    "LISAconst": "Proposal",
    "responseapprox": "full",
    "frozenLISA": False,
    "TDIrescaled": True,
    "LISAnoise": {
    "InstrumentalNoise": "SciRDv1",
    "WDbackground": True,
    "WDduration" : 3.0
    },
    "waveform_params_update_for_template": {}
}

# Default pymultinest params
# TODO: print_info, n_iter_info very rudimentary at the moment
run_params_musthave = ['out_dir', 'out_name']
run_params_default = {
    "sampler": "ptemcee",
    "sample_Lframe": True,
    "multimodal": False,
    "multimodal_pattern": "1mode",
    "p_jump": 0.,
    "params_map": None,
    "ensemble_proposal": "ptemcee",
    "likelihood_method": "fresnel",
    "likelihood_residuals_ngrid": None,
    "skip_fisher": False,
    "init_method": "prior",
    "init_file": None,
    "init_scale_cov": 1.,
    "zerolike": False,
    "n_temps": 5,
    "temp_max": None,
    "adaptive_temp": False,
    "n_walkers": 100,
    "n_iter": 1000,
    "burn_in": 100,
    "autocor_method": "acor",
    "thin_samples": True,
    "upsample": 1,
    "seed": None,
    "print_info": False,
    "n_iter_info": 10,
    "output": True,
    "output_raw": True
}

# Default parameter range for angular parameters
params_range_default = {
    "inc": [0., np.pi],
    "phi": [-np.pi, np.pi],
    "lambda": [-np.pi, np.pi],
    "beta": [-np.pi/2, np.pi/2],
    "psi": [0., np.pi]
}


################################################################################
# Main

if __name__ == '__main__':

    # Using MPI or not
    # We allow for different cases: MPI used or not, master or worker
    use_mpi = False
    is_master = True
    mapper = map
    if MPI is not None:
        MPI_size = MPI.COMM_WORLD.Get_size()
        MPI_rank = MPI.COMM_WORLD.Get_rank()
        use_mpi = (MPI_size > 1)
        if use_mpi:
            print("MPI rank/size: %d / %d" % (MPI_rank, MPI_size), flush=True)
            pool = ptemcee.mpi_pool.MPIPool(debug=False)
            is_master = pool.is_master()
            mapper = pool.map
        else:
            print("No MPI", flush=True)
            is_master = True
            mapper = map

    # Load json file with all parameters
    with open(args.input_file, 'r') as input_file:
        input_params = json.load(input_file)
    # Source params and prior params must be given
    source_params = input_params['source_params']
    source_params = pytools.complete_mass_params(source_params)
    prior_params = input_params['prior_params']
    # Source params can be specified in the Lframe
    source_params_are_Lframe = source_params.get('Lframe', False)
    # Waveform params and run params have predefined default values
    # Need to cast from json to python, list of modes list[list] -> list[tuple]
    waveform_params = waveform_params_default.copy()
    input_params['waveform_params'] = inference.waveform_params_json2py(
                                                input_params['waveform_params'])
    waveform_params.update(input_params['waveform_params'])
    run_params = run_params_default.copy()
    run_params.update(input_params['run_params'])
    zerolike = run_params.pop('zerolike', False)
    sample_Lframe = run_params.pop('sample_Lframe', False)
    simple_likelihood = run_params.pop('simple_likelihood', False)

    # Only master will print info
    print_info = run_params['print_info'] and is_master

    # Check run params are meant to be for the ptemcee sampler
    if not (run_params['sampler'] == 'ptemcee'):
        raise ValueError('run_params sampler flag is %s instead of ptemcee.'\
                          % (run_params['sampler']))
    # run params must contain output prefix for files -- no default here
    for musthave in run_params_musthave:
        if not (musthave in run_params):
            raise ValueError('run_params must contain %s.' % (musthave))
    # Check that the output directory exists
    if (run_params['output'] or run_params['output_raw']):
        if not os.path.isdir(run_params['out_dir']):
            raise ValueError('Output dir %s does not exist.' % run_params['out_dir'])

    # If source params are given in the Lframe, convert to SSB
    if source_params_are_Lframe:
        source_params_SSBframe = lisatools.convert_Lframe_to_SSBframe(
                                    source_params,
                                    t0=waveform_params['t0'],
                                    frozenLISA=waveform_params['frozenLISA'])
        source_params_Lframe = source_params.copy()
    else:
        source_params_SSBframe = source_params.copy()
        source_params_Lframe = lisatools.convert_SSBframe_to_Lframe(
                                    source_params,
                                    t0=waveform_params['t0'],
                                    frozenLISA=waveform_params['frozenLISA'])
    # Complete mass and spin injection parameters
    source_params_SSBframe = pytools.complete_mass_params(source_params_SSBframe)
    source_params_SSBframe = pytools.complete_spin_params(source_params_SSBframe)
    source_params_Lframe = pytools.complete_mass_params(source_params_Lframe)
    source_params_Lframe = pytools.complete_spin_params(source_params_Lframe)

    # Use default param ranges to complete input ranges -- check completeness
    prior_type = prior_params['prior_type']
    params_range = prior_params['params_range']
    infer_params = prior_params['infer_params']
    for i, param in enumerate(infer_params):
        if params_range[i] == []:
            params_range[i] = params_range_default[param]
    if [] in params_range:
        raise ValueError('Some parameter ranges are neither given \
                      nor covered by the defaults.')

    # Parameters inferred and parameters fixed
    list_params = prior_params.get('list_params', list_params_default)
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in list_params if p not in infer_params]
    n_dim = len(infer_params)
    n_fixed = len(fixed_params)
    # Vector x of infer_params for the injection
    # Also store values of fixed params in dict for later use
    if sample_Lframe:
        fixed_params_dict = dict([(p, source_params_Lframe[p]) for p in fixed_params])
        x_inj = np.array([source_params_Lframe[p] for p in infer_params])
    else:
        fixed_params_dict = dict([(p, source_params_SSBframe[p]) for p in fixed_params])
        x_inj = np.array([source_params_SSBframe[p] for p in infer_params])

    # Wrapped parameters
    wrap_params = prior_params.get('wrap_params', None)
    if wrap_params is None:
        wrap_params = inference.func_default_wrap(infer_params)
    else:
        if np.array(wrap_params).dtype == 'bool':
            raise ValueError('wrap_params in outdated format, expecting None (for default) or a list of [None or [a,b]] with [a,b] the wrap moduli, typically [-np.pi, np.pi].')
    # When using a parameter map, wrap_params is given for the mapped params
    # Length must match n_dim_ensemble, see below the part about maps

    # Ensemble proposal:
    # ~1/sqrt(z) with volume element z^(d-1) for emcee
    # ~1/z with volume element z^d for ptemcee
    ensemble_proposal = run_params['ensemble_proposal']

    # Multimodal jumps
    # TODO: support both 2-mode and 8-mode jumps
    # TODO: be more flexible when some parameters are pinned to their value
    # For now we require all ['inc', 'lambda', 'beta', 'psi'] to be
    # inferred params
    # TODO: improve the back-and-forth between vector and dict
    multimodal = run_params['multimodal']
    multimodal_pattern = run_params['multimodal_pattern']
    extra_proposal_prob = run_params['p_jump']
    extra_proposal_jump = None
    if multimodal:
        if not sample_Lframe:
            raise ValueError('The multimodal option is only supported \
                                together with sample_Lframe.')

        list_angle_params = ['inc', 'lambda', 'beta', 'psi']
        index_map_angles = {}
        for param in list_angle_params:
            if not param in infer_params:
                raise ValueError('For multimodality, infer_params must contain \
                                        [inc, phi, lambda, beta, psi]')
            index_map_angles[param] = infer_params.index(param)
        # x is (n, ndim) array with physical parameters (for n walkers)
        # returns x_jump
        def jump_extra(x, random_state=np.random):
            angle_params = {}
            for param in list_angle_params:
                angle_params[param] = x[:,index_map_angles[param]]
            skymodes = inference.proposal_skymode(size=len(x), pattern=multimodal_pattern, random_state=random_state)
            angle_params_map = inference.map_skymode(angle_params, skymodes[:,0], skymodes[:,1])
            x_jump = x.copy()
            for param in list_angle_params:
                x_jump[:,index_map_angles[param]] = angle_params_map[param]
            return x_jump
        extra_proposal_jump = jump_extra

    # Prior function (assumes independent multiplicative prior in all params)
    # prior_funcs is a list of pure lambda functions
    prior_funcs = []
    for i in range(n_dim):
        prior_funcs += [inference.def_prior_func(prior_type[i], params_range[i])]
    def prior(x):
        p = 1.
        for i in range(n_dim):
            p *= prior_funcs[i](x[i])
        return p
    # Set up prior bounds for x vector of parameters, dimension (n_dim, 2)
    prior_bounds = np.array(params_range)
    # Check the injection is within the prior bounds
    if not inference.prior_check(x_inj, prior_bounds):
        print("x_inj:")
        print(x_inj)
        print("prior_bounds:")
        print(prior_bounds)
        raise ValueError('Source parameters not within the prior bounds!')

    # Set up likelihood class
    if run_params['likelihood_method']=="simple_likelihood":
        #TODO: check if all options are right (intrinsic params pinned)
        likelihoodClass = lisa.SimpleLikelihoodLISASMBH(source_params_Lframe,
                                                        **waveform_params)
    elif run_params['likelihood_method']=="fresnel":
        likelihoodClass = lisa.LikelihoodLISASMBH(source_params_Lframe,
                                                  **waveform_params)
    elif run_params['likelihood_method']=="residuals":
        ngrid = run_params['likelihood_residuals_ngrid']
        likelihoodClass = lisa.LikelihoodLISASMBH_LinearResiduals(
                                                  source_params_Lframe,
                                                  ngrid=ngrid,
                                                  **waveform_params)
    else:
        raise ValueError('Likelihood method %s not recognized.' \
                                            % run_params['likelihood_method'])

    # Waveform params to be used for template waveforms
    # May differ from the waveform params for injection, given as an update
    waveform_params_template = copy.deepcopy(waveform_params)
    if 'waveform_params_update_for_template' in waveform_params:
        waveform_params_template.update(
                         waveform_params['waveform_params_update_for_template'])

    # Option to use a parameter map inspired from the simple 22-mode response
    # Sampling extrinsic params (see (74) in arXiv:2003.00357):
    # A_\pm exp(i Phi_\pm) = sigma^22_\pm
    # Set of params: (A_+, A_-, Phi_+ - Phi_-, Phi_+ + Phi_-, lambdaL, betaL)
    params_map = run_params['params_map']
    if params_map is None: # When not using a parameter map
        n_dim_map = n_dim
        n_dim_ensemble = n_dim
        def map_params(x):
            return x
        def invmap_params_and_jacobian(x):
            return x, 1.
        # When not using a map, prior ranges in original parameters take care of
        # checking the range
        def physical_check_map(x):
            return True
    elif params_map=='simple_response_22':
        if print_info:
            print("Using params_map: simple_response_22", flush=True)
        # Indices of extrinsic parameters, must sample_Lframe, have the full set
        if not sample_Lframe:
            raise ValueError('sample_simple22_map requires sample_Lframe=True.')
        list_params_extr = ['dist', 'inc', 'phi', 'lambda', 'beta', 'psi']
        list_params_extr_map = ['Aplus', 'Aminus', 'Phiplus', 'Phiminus', 'lambda', 'sbeta', 'indexpsi']
        map_wrap_dict = {
            'Phiplus': [-np.pi, np.pi],
            'Phiminus': [-np.pi, np.pi],
            'lambda': [-np.pi, np.pi]
            }
        indices_extr = np.zeros(6, dtype=int)
        for i,p in enumerate(list_params_extr):
            indices_extr[i] = infer_params.index(p)
        indices_intr = np.array([infer_params.index(p) for p in infer_params if not p in list_params_extr])
        # Place first 6 map params where the original extrinsic params were
        # This is arbitrary, just making sure we don't touch intrinsic params
        indices_extr_map = np.zeros(7, dtype=int)
        indices_extr_map[:6] = indices_extr
        # We require an extra discrete param, indexpsi=0,1, for the map
        # Will be placed last in extended x used for internal sampling
        # The ensemble proposal with still only use the first n_dim
        indices_extr_map[6] = n_dim
        n_dim_map = n_dim + 1
        n_dim_ensemble = n_dim
        # List of infer_params mapped
        infer_params_map = infer_params.copy()
        for i in range(6):
            infer_params_map[indices_extr[i]] = list_params_extr_map[i]
        infer_params_map += ['indexpsi']
        # We need to adjust extra_proposal to work with the mapped parameters
        if multimodal:
            list_angle_params = ['lambda', 'sbeta', 'indexpsi']
            indices_map_angles = [list_params_extr_map.index(p) for p in list_angle_params]
            def jump_extra(x, random_state=np.random):
                angle_params = {}
                for i,p in enumerate(list_angle_params):
                    angle_params[p] = x[:,indices_extr_map[indices_map_angles[i]]]
                skymodes = inference.proposal_skymode_simple22map(size=len(x), pattern=multimodal_pattern, random_state=random_state)
                angle_params_map = inference.map_skymode_simple22map(angle_params, skymodes[:,0], skymodes[:,1], skymodes[:,2])
                x_jump = x.copy()
                for i,p in enumerate(list_angle_params):
                    x_jump[:,indices_extr_map[indices_map_angles[i]]] = angle_params_map[p]
                return x_jump
            extra_proposal_jump = jump_extra
        # When using a parameter map, wrap_params is given for the mapped params
        # Length must match n_dim_ensemble
        # If None was given as input (default), set the wrap automatically
        if prior_params.get('wrap_params', None) is None:
            list_wrap_moduli = []
            for i in range(n_dim_ensemble):
                list_wrap_moduli += [map_wrap_dict.get(infer_params_map[i], None)]
            # Overriding previous default initialization when input is None
            wrap_params = list_wrap_moduli
        # Functions working with a vector x
        injdist = source_params['dist']
        # Function to check the physical range of mapped params
        def physical_check_map(x):
            params_map = {}
            for i,p in enumerate(list_params_extr_map):
                params_map[p] = x[indices_extr_map[i]]
            return inference.physical_check_simple_response_22(params_map)
        # Map from ordinary params x to mapped internal params X
        def map_params(x):
            params = {}
            for i,p in enumerate(list_params_extr):
                params[p] = x[indices_extr[i]]
            params_map = inference.map_params_simple_response_22(params, injdist)
            x_map = np.zeros(len(x)+1, dtype=float) # Add a slot for indexpsi
            x_map[indices_intr] = x[indices_intr]
            for i,p in enumerate(list_params_extr_map):
                x_map[indices_extr_map[i]] = params_map[p]
            return x_map
        # Inverse map from mapped internal params X to ordinary params x
        # Also compute jacobian J = |\partial X / \partial x|
        def invmap_params_and_jacobian(x):
            params_map = {}
            for i,p in enumerate(list_params_extr_map):
                params_map[p] = x[indices_extr_map[i]]
            params = inference.invmap_params_simple_response_22(params_map, injdist)
            jacobian = inference.jacobian_map_simple_response_22(params, params_map, injdist)
            x_invmap = np.zeros(len(x)-1, dtype=float) # Cutting the slot added for indexpsi
            x_invmap[indices_intr] = x[indices_intr]
            for i,p in enumerate(list_params_extr):
                x_invmap[indices_extr[i]] = params[p]
            return x_invmap, jacobian
    else:
        raise ValueError('params_map %s not recognized.' % params_map)

    # Define ln-likelihood and ln-prior function, wrapped for use in ptemcee
    # TODO: add m1>m2 to prior range checks
    def lnlikelihood(x):
        # Allowing to use a parameter map, like for simple22_map
        # Check physical range in the possibly mapped params
        if not physical_check_map(x):
            return -1e99
        x_invmap, jacobian_map = invmap_params_and_jacobian(x)
        # Check prior range
        if not inference.prior_check(x_invmap, prior_bounds):
            return -1e99
        # Convert to dictionary of params
        template_params = {}
        for i in range(n_dim):
            template_params[infer_params[i]] = x_invmap[i]
        for p in fixed_params:
            template_params[p] = fixed_params_dict[p]
        template_params = pytools.complete_mass_params(template_params)
        template_params = pytools.complete_spin_params(template_params)
        template_params['Lframe'] = sample_Lframe
        if zerolike: return 0.
        lnL = likelihoodClass.lnL(template_params, **waveform_params_template)
        return lnL
    def lnprior(x):
        # Allowing to use a parameter map, like for simple22_map
        # Check physical range in the possibly mapped params
        if not physical_check_map(x):
            return -1e99
        x_invmap, jacobian_map = invmap_params_and_jacobian(x)
        if not inference.prior_check(x_invmap, prior_bounds):
            return -1e99
        p = prior(x_invmap)
        if p==0.: return -1e99
        lnp = np.log(p)
        # Include effect of the (inverse) jacobian of the parameter map, if any
        lnjacobian = -np.log(np.abs(jacobian_map))
        return lnp + lnjacobian

    # Read run_params
    # p_extra = run_params['p_jump']
    init_method = run_params['init_method']
    init_scale_cov = run_params['init_scale_cov']
    n_temps = run_params['n_temps']
    temp_max = run_params['temp_max']
    adaptive_temp = run_params['adaptive_temp']
    n_walkers = run_params['n_walkers']
    n_iter = run_params['n_iter']
    burn_in = run_params['burn_in']
    autocor_method = run_params['autocor_method']
    upsample = run_params['upsample']
    seed = run_params['seed']
    n_iter_info = run_params['n_iter_info']
    n_dim = len(infer_params)

    if print_info:
        print('Source parameters SSBframe:')
        print(source_params_SSBframe)
        print('Source parameters Lframe:')
        print(source_params_Lframe)

    # Random seed for the sampling
    if seed is not None:
        np.random.seed(seed)

    # Compute Fisher matrix - not available for all waveform models
    if not run_params['skip_fisher']:
        # NOTE: might be slow if timetomerger_max is set at generic value like 1yr
        # NOTE: numerical problems in Fisher when timetomerger_max is too large
        # For Fisher only, we set timetomerger_max based on SNR=1 threshold
        waveform_params_22 = waveform_params.copy()
        waveform_params_22['modes'] = [(2,2)]
        t_SNR1_s = snrtools.lisa_tofSNR(1., source_params_SSBframe, **waveform_params_22)
        t_SNR1_yr = t_SNR1_s / pyconstants.YRSID_SI
        # Ensure at least 1d of signal for fisher regardless of SNR threshold
        t_threshold_yr = np.fmax(t_SNR1_yr, 1./365.25)
        waveform_params_fisher = waveform_params.copy()
        waveform_params_fisher['timetomerger_max'] = t_threshold_yr
        # TODO: improve handling of different choices of infer_params
        if print_info:
            print('Computing Fisher matrix...')
            print('Threshold max(t(SNR=1), 1d) used for Fisher: %g yr' % (t_threshold_yr))
        # Computing Fisher matrix and covariance
        fishercov = lisa_fisher.fisher_covariance_smbh(source_params_SSBframe,
                            steps=lisa_fisher.default_steps, freqs=['log', None],
                            list_params=infer_params, list_fixed_params=fixed_params,
                            Lframe=sample_Lframe,
                            **waveform_params_fisher)
        if print_info:
            print('Fisher covariance matrix:')
            print(fishercov['cov'])

    # Initialize walkers, drawing from the prior or from Fisher matrix
    x_ini = np.zeros((n_temps, n_walkers, n_dim))
    if init_method=='prior':
        # Drawing from the prior -- assumes independent priors in all params
        for k in range(n_temps):
            for j in range(n_dim):
                x_ini[k,:,j] = inference.draw_prior(prior_type[j], prior_bounds[j], n_walkers)
    elif init_method=='fisher':
        if run_params['skip_fisher']:
            raise ValueError('Cannot have skip_fisher and init_method=fisher.')
        # Drawing from the Fisher matrix - same for each temperature
        if sample_Lframe:
            mean = np.array([source_params_Lframe[p] for p in infer_params])
        else:
            mean = np.array([source_params_SSBframe[p] for p in infer_params])
        # We allow for the possibility of scaling the covariance for initialization
        cov = fishercov['cov'] * init_scale_cov
        try:
            for k in range(n_temps):
                for i in range(n_walkers):
                    x_ini[k,i,:] = inference.draw_multivariate_normal_constrained(
                                          mean, cov, params_range, infer_params)
        except:
            # Regularizing Fisher with a prior inverse covariance in spins
            print('Fisher initialization failed, adding regularizing prior invcov at sigma=0.5 for spins.')
            # By default, add a Gaussian prior with sigma=0.5 in both spins
            fisher_prior_invcov = np.zeros((n_dim,n_dim), dtype=float)
            if 'chi1' in infer_params:
                ichi1 = infer_params.index('chi1')
                fisher_prior_invcov[ichi1,ichi1] = 1./0.5**2
            if 'chi2' in infer_params:
                ichi2 = infer_params.index('chi2')
                fisher_prior_invcov[ichi2,ichi2] = 1./0.5**2
            fishercov_reg = lisa_fisher.fisher_covariance_smbh(source_params_SSBframe,
                                steps=lisa_fisher.default_steps, freqs=['log', None],
                                list_params=infer_params, list_fixed_params=fixed_params,
                                Lframe=sample_Lframe, prior_invcov=fisher_prior_invcov,
                                **waveform_params_fisher)
            print('Fisher covariance matrix, regularized in spins:')
            print(fishercov_reg['cov'])
            cov_reg = fishercov_reg['cov'] * init_scale_cov
            try:
                for k in range(n_temps):
                    for i in range(n_walkers):
                        x_ini[k,i,:] = inference.draw_multivariate_normal_constrained(
                                              mean, cov_reg, params_range, infer_params)
            except:
                print('Fisher initialization failed with regularization, initializing from prior.')
                # Drawing from the prior -- assumes independent priors in all params
                for k in range(n_temps):
                    for j in range(n_dim):
                        x_ini[k,:,j] = inference.draw_prior(prior_type[j], prior_bounds[j], n_walkers)
    elif init_method=='file':
        init_file = run_params['init_file']
        with h5py.File(init_file, 'r') as hf:
            # NOTE: hf['Lframe'] is expected as a dataset of just one element
            if (hf['Lframe'][0] != sample_Lframe):
                raise ValueError('sample_Lframe and Lframe in posterior loaded from file must match.')
            for j in range(n_dim):
                p = infer_params[j]
                x_ini[:,:,j] = np.reshape(hf[p][:], (n_temps, n_walkers))
    else:
        raise ValueError('Option init_method = %s not recognized.' % init_method)

    # Initial points all start on the main mode (1,0)
    # if multimodal:
    #     x_ini[:,n_dim] = 1
    #     x_ini[:,n_dim+1] = 0

    # Enforce range of phase parameters in case Fisher uncertainties are large
    # TODO: we should have proper rejection based on the prior ranges
    if 'inc' in infer_params:
        iinc = infer_params.index('inc')
        x_ini[:,:,iinc] = pytools.modpi(x_ini[:,:,iinc])
    if 'phi' in infer_params:
        iphi = infer_params.index('phi')
        x_ini[:,:,iphi] = pytools.mod2pi(x_ini[:,:,iphi])
    if 'lambda' in infer_params:
        ilambda = infer_params.index('lambda')
        x_ini[:,:,ilambda] = pytools.mod2pi(x_ini[:,:,ilambda])
    if 'beta' in infer_params:
        ibeta = infer_params.index('beta')
        x_ini[:,:,ibeta] = pytools.modpi2(x_ini[:,:,ibeta])
    if 'psi' in infer_params:
        ipsi = infer_params.index('psi')
        x_ini[:,:,ipsi] = pytools.mod2pi(x_ini[:,:,ipsi])

    # When using a parameter map, map the initial x
    x_ini_map = np.zeros((n_temps, n_walkers, n_dim_map), dtype=float)
    for i in range(n_temps):
        for j in range(n_walkers):
            x_ini_map[i,j,:] = map_params(x_ini[i,j,:])

    # Temperature ladder
    betas = ptemcee.sampler.make_ladder(n_dim, ntemps=n_temps, Tmax=temp_max)

    # If using mpi, workers are simply told to wait
    # They will be sollicited by master for likelihood computations via pool.map
    if use_mpi and (not is_master):
        pool.wait()
        # Is that exit necessary ? Was in the emcee/schwimmbad example
        # Maybe there to ensure exits if pool.close() is not called
        sys.exit(0)

    # Whether using mpi and being master or simply not using mpi
    # Sampler set-up, running, post-processing and output all done by master
    if is_master:
        # Set up and initialize sampler
        sampler = ptemcee.sampler.Sampler(n_walkers, n_dim_map,
                                          lnlikelihood, lnprior,
                                          ndim_ensemble=n_dim_ensemble,
                                          betas=betas, mapper=mapper,
                                          adaptive=adaptive_temp,
                                          extra_proposal_prob=extra_proposal_prob,
                                          extra_proposal_jump=extra_proposal_jump,
                                          ensemble_proposal=ensemble_proposal,
                                          list_param_wrap=wrap_params)
        chain = sampler.chain(x_ini_map)

        # A la grace de dieu
        if print_info:
            print('Running ptemcee...')
        chain.run(n_iter)

        # TODO: output meta information, acceptance rate and likelihood speed

        # Post-processing: unfold if multimodal was used
        if print_info:
            print('Post-processing...')

        # Only keep 0-temperature chain for post-processing
        # TODO: allow to output other temperatures, and temp. adjustments, for debug
        # ptemcee format: (niter, ntemp, nwalker, ndim)
        # We translate to multiemcee format: (nwalker, niter, ndim) for temp=0
        # NOTE: when using a map, chain.x has mapped values and dimension
        chain_unfold_map = np.transpose(chain.x[:,0,:,:], axes=(1,0,2))
        # Same format without ndim for like and post values
        lnlikevals = np.transpose(chain.logl[:,0,:], axes=(1,0))
        lnpostvals = np.transpose(chain.logP[:,0,:], axes=(1,0))
        # Format: (niter, ntemp)
        betavals = chain.betas
        # ptemcee does not seem to store individual acceptance, only their count
        acceptance_ratios = chain.jump_acceptance_ratio[0,:]

        # When using a map, do the inverse mapping for the chain's x values
        chain_unfold = np.zeros((n_walkers, n_iter, n_dim), dtype=float)
        for i in range(n_walkers):
            for j in range(n_iter):
                chain_unfold[i,j,:], _ = invmap_params_and_jacobian(chain_unfold_map[i,j,:])

        # Post-processing: burn-in, autocor. length and thinning, merge walkers
        # Also complete with params that were fixed to their values
        if burn_in>=n_iter:
            print('WARNING: burn_in is larger than n_iter, ignoring burn_in.')
            burn_in = 0
        if run_params['thin_samples']:
            autocor_len = {}
            for i,param in enumerate(infer_params):
                # Note: autocorr_new computes a mean across walkers
                autocor_len[param] = inference.get_autocor(chain_unfold[:,burn_in:,i], method=autocor_method)
                if print_info:
                    print('Autocorrelation length for %s: %g' % (param,autocor_len[param]))
            # The thinning length is the worse autocor length for all params
            thin_len = int(np.ceil(np.max([autocor_len[param] for param in infer_params])))
            thin_len = int(np.ceil(thin_len / float(upsample)))
        else:
            thin_len = 1
        if print_info:
            print('Autocorrelation thinning: %d' % (thin_len,))
        # Processed output: thinned and merged chain (completed with fixed params)
        n_out = n_walkers * ((n_iter-burn_in-1) // thin_len + 1)
        chain_processed = np.zeros((n_out, n_dim + n_fixed), dtype=float)
        # Here ravel with order='F' flattens by reading along columns first
        for i in range(n_dim):
            i_out = list_params.index(infer_params[i])
            chain_processed[:,i_out] = np.ravel(chain_unfold[:,burn_in::thin_len,i],
                                                order='F')
        for i in range(n_fixed):
            i_out = list_params.index(fixed_params[i])
            if sample_Lframe:
                chain_processed[:,i_out] = source_params_Lframe[fixed_params[i]]
            else:
                chain_processed[:,i_out] = source_params_SSBframe[fixed_params[i]]
        # Thin and merge likelihood, posterior, acceptance vals
        lnlikevals_processed = np.ravel(lnlikevals[:,burn_in::thin_len], order='F')
        lnpostvals_processed = np.ravel(lnpostvals[:,burn_in::thin_len], order='F')
        # acceptvals_processed = np.ravel(acceptvals[:,burn_in::thin_len], order='F')

        # Output
        basename = run_params['out_dir'] + run_params['out_name']
        # Main output: processed, thinned chains and Fisher matrix
        # TODO: hierarchy, create a level (group ?) samples
        if run_params['output']:
            with h5py.File(basename + '.h5', 'w') as f:
                # Source params
                source_params_Lframe_gr = f.create_group('source_params_Lframe')
                pytools.write_h5py_dict(source_params_Lframe_gr, source_params_Lframe)
                source_params_SSBframe_gr = f.create_group('source_params_SSBframe')
                pytools.write_h5py_dict(source_params_SSBframe_gr, source_params_SSBframe)
                for i,param in enumerate(list_params):
                    f.create_dataset(param, data=chain_processed[:,i])
                f.create_dataset('lnlike', data=lnlikevals_processed)
                f.create_dataset('lnpost', data=lnpostvals_processed)
                f.create_dataset('acceptance_ratios', data=acceptance_ratios)
                f.create_dataset('betas', data=betavals)
                # f.create_dataset('accept', data=acceptvals_processed)
                if not run_params['skip_fisher']:
                    fishercov_gr = f.create_group('fishercov')
                    lisa_fisher.write_h5py_fishercov(fishercov_gr, fishercov)
        # Extra output: full chains, walker-separated, for testing and diagnostics
        if run_params['output_raw']:
            with h5py.File(basename + '_raw.h5', 'w') as f:
                for i,param in enumerate(infer_params):
                    f.create_dataset(param, data=chain_unfold[:,:,i])
                f.create_dataset('lnlike', data=lnlikevals)
                f.create_dataset('lnpost', data=lnpostvals)
                # f.create_dataset('accept', data=acceptvals)

        # If using mpi, close the pool
        if use_mpi:
            pool.close()



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Python functions for Fisher matrix computations for LISA.
"""


import numpy as np
import copy

import lisabeta
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools
import lisabeta.tools.pyspline as pyspline
import lisabeta.tools.pyoverlap as pyoverlap
import lisabeta.lisa.pyresponse as pyresponse
import lisabeta.lisa.pyLISAnoise as pyLISAnoise
import lisabeta.lisa.lisatools as lisatools
import lisabeta.lisa.lisa as lisa
import lisabeta.waveforms.bbh.pyIMRPhenomD as pyIMRPhenomD
import lisabeta.waveforms.bbh.pyIMRPhenomHM as pyIMRPhenomHM
import lisabeta.waveforms.bbh.pyEOBNRv2HMROM as pyEOBNRv2HMROM

default_list_fisher_params = ['M', 'q', 'chi1', 'chi2', 'Deltat', 'dist', 'inc', 'phi', 'lambda', 'beta', 'psi']

dict_params_fractional = {
    'M':True,
    'Mchirp':True,
    'm1':True,
    'm2':True,
    'q':False,
    'eta':False,
    'chi1':False,
    'chi2':False,
    'chip':False,
    'chim':False,
    'chis':False,
    'chia':False,
    'chiPN':False,
    'Deltat':False,
    'fstart':False,
    'dist':True,
    'inc':False,
    'phi':False,
    'lambda':False,
    'beta':False,
    'psi':False
}

default_steps = {
    'M':2e-8,
    'Mchirp':1e-8,
    'm1':2e-6,
    'm2':2e-6,
    'q':1e-7,
    'eta':1e-7,
    'chi1':1e-7,
    'chi2':1e-7,
    'chip':1e-7,
    'chim':1e-7,
    'chis':1e-7,
    'chia':1e-7,
    'chiPN':1e-7,
    'Deltat':1e-4,
    'fstart':1e-10,
    'dist':2e-8,
    'inc':1e-7,
    'phi':1e-7,
    'lambda':1e-7,
    'beta':1e-7,
    'psi':1e-7
}

def complete_params_dict(params):
    params_ext = params.copy()
    m1 = params['m1']
    m2 = params['m2']
    params_ext['M'] = m1 + m2
    params_ext['q'] = m1 / m2
    return params_ext

def fisher_element(params, param1, param2, step1, step2, freqs, Lframe=False, **waveform_params):
    if Lframe:
        raise ValueError('Lframe not implemented yet, sorry.')

    # Parameters
    m1 = params['m1']
    m2 = params['m2']
    M = m1 + m2
    q = m1 / m2
    params['M'] = M
    params['q'] = q

    LISAnoise = waveform_params.pop('LISAnoise', pyLISAnoise.LISAnoiseSciRDv1)
    TDI = waveform_params.get('TDI', 'TDIAET')
    TDIrescaled = waveform_params.get('TDIrescaled', True)
    LISAconst = waveform_params.get('LISAconst', pyresponse.LISAconstProposal)

    # Parameter steps
    frac1 = dict_params_fractional[param1]
    frac2 = dict_params_fractional[param2]
    params_base = params.copy()
    params_step_1 = params.copy()
    params_step_2 = params.copy()
    if frac1:
        params_step_1[param1] *= 1 + step1
    else:
        params_step_1[param1] += step1
    if frac2:
        params_step_2[param2] *= 1 + step2
    else:
        params_step_2[param2] += step2
    M1 = params_step_1['M']
    q1 = params_step_1['q']
    M2 = params_step_2['M']
    q2 = params_step_2['q']
    params_step_1['m1'] = M1 * q1 / (1+q1)
    params_step_1['m2'] = M1 * 1 / (1+q1)
    params_step_2['m1'] = M2 * q2 / (1+q2)
    params_step_2['m2'] = M2 * 1 / (1+q2)

    # Generate tdi freqseries
    tdifreqseries_base = lisa.GenerateLISATDIFreqseries_SMBH(params_base, freqs, **waveform_params)
    tdifreqseries_step_1 = lisa.GenerateLISATDIFreqseries_SMBH(params_step_1, freqs, **waveform_params)
    tdifreqseries_step_2 = lisa.GenerateLISATDIFreqseries_SMBH(params_step_2, freqs, **waveform_params)

    # Compute noises
    noise_evaluator = pyLISAnoise.initialize_noise(LISAnoise,
                                        TDI=TDI, TDIrescaled=TDIrescaled,
                                        LISAconst=LISAconst)
    Sn1_vals, Sn2_vals, Sn3_vals = pyLISAnoise.evaluate_noise(
                          LISAnoise, noise_evaluator, freqs,
                          TDI=TDI, TDIrescaled=TDIrescaled, LISAconst=LISAconst)

    # Compute derivatives d1h, d2h
    modes = tdifreqseries_base['modes']
    channels = ['chan1', 'chan2', 'chan3']
    Snvals = {}
    Snvals['chan1'] = Sn1_vals
    Snvals['chan2'] = Sn2_vals
    Snvals['chan3'] = Sn3_vals
    d1h = {}
    d2h = {}
    for chan in channels:
        d1h[chan] = np.zeros_like(freqs, dtype=complex)
        d2h[chan] = np.zeros_like(freqs, dtype=complex)
        for lm in modes:
            d1h[chan] += (tdifreqseries_step_1[lm][chan] - tdifreqseries_base[lm][chan]) / step1
            d2h[chan] += (tdifreqseries_step_2[lm][chan] - tdifreqseries_base[lm][chan]) / step2

    # Compute (d1h|d2h)
    # We do not assume that deltaf is constant
    df = np.diff(freqs)
    Fij = 0.
    for chan in channels:
        Fij += 4*np.sum(df * np.real(d1h[chan] * np.conj(d2h[chan]) / Snvals[chan])[:-1])

    return Fij

def fisher_covariance_smbh(params, freqs=None, steps=default_steps, list_params=default_list_fisher_params, list_fixed_params=[], Lframe=False, prior_invcov=None, fLow_from_22=True, **waveform_params):
    '''
    This updated version of Fisher computation:
      1. Adds options for additional mass parameter combinations
      2. Adds option for providing a "prior" inverse covariance which is added to the Fisher matrix
      3. Default is as before
    '''

    waveform_params = copy.deepcopy(waveform_params)

    # Parameters
    Npar = len(list_params)

    LISAnoise = waveform_params.get('LISAnoise', pyLISAnoise.LISAnoiseSciRDv1)
    TDI = waveform_params.get('TDI', 'TDIAET')
    TDIrescaled = waveform_params.get('TDIrescaled', True)
    LISAconst = waveform_params.get('LISAconst', pyresponse.LISAconstProposal)

    # Convert input parameters to either Lframe or SSBframe
    if not params.get('Lframe', False) and Lframe:
        params_base = lisatools.convert_SSBframe_to_Lframe(
                            params,
                            t0=waveform_params['t0'],
                            frozenLISA=waveform_params['frozenLISA'])
    elif params.get('Lframe', False) and not Lframe:
        params_base = lisatools.convert_Lframe_to_SSBframe(
                            params,
                            t0=waveform_params['t0'],
                            frozenLISA=waveform_params['frozenLISA'])
    else:
        params_base = params.copy()

    # Masses and spins
    params_base = pytools.complete_mass_params(params_base)
    params_base = pytools.complete_spin_params(params_base)
    m1 = params_base['m1']
    m2 = params_base['m2']
    M = params_base['M']
    Ms = M * pyconstants.MTSUN_SI

    channels = ['chan1', 'chan2', 'chan3']

    # Generate base waveform, with frequency bounds from waveform_params
    # e.g. scale_freq_hm might be True or False
    wftdi_base = lisa.GenerateLISATDI_SMBH(params_base, **waveform_params)

    # Read set of modes and dict of mode-by-mode freq bounds from base waveform
    modes = wftdi_base['modes']
    fLow_lm = dict([(lm,wftdi_base[lm]['freq'][0]) for lm in modes])
    fHigh_lm = dict([(lm,wftdi_base[lm]['freq'][-1]) for lm in modes])
    # Option to ignore low freqs where only 21 has support
    # This 21-specific frequency range is costly to cover with 22-based grid
    # but normally low SNR
    if fLow_from_22:
        fLow = fLow_lm[(2,2)]
    else:
        fLow = np.min([fLow_lm[lm] for lm in modes])
    fHigh = np.max([fHigh_lm[lm] for lm in modes])

    # Default frequencies (safe and slow):
    # linear spacing, adjust deltaf=1/(2T) with T approximate duration
    if freqs is None:
        freqs = ['linear', None]
    # Determine the frequencies to use for the overlaps, if not given as input
    if not isinstance(freqs, np.ndarray):

        # Format: freqs = ['log', npt], npt=None for auto setting with T
        if freqs[0]=='log':
            if freqs[1] is not None:
                nptlog = int(freqs[1])
            else:
                # Estimate length and choose deltaf according to Nyquist
                # NOTE: No safety margin, so set timetomerger_max must carefully
                T = pytools.funcNewtoniantoff(m1, m2, fLow)
                deltaf = 1./(2*T)
                nptlog = np.ceil(np.log(fHigh/fLow) / np.log(1. + deltaf/fLow))
            freqs = pytools.logspace(fLow, fHigh, nptlog)

        # Format: freqs = ['linear', deltaf], deltaf=None for auto setting with T
        if freqs[0]=='linear':
            if freqs[1] is not None:
                deltaf = freqs[1]#int(freqs[1])
            else:
                # Estimate length and choose deltaf according to Nyquist
                # NOTE: No safety margin, so set timetomerger_max must carefully
                T = pytools.funcNewtoniantoff(m1, m2, fLow)
                deltaf = 1./(2*T)
            freqs = np.arange(fLow, fHigh, deltaf)

        # Format: freqs = ['nyquist_log', pars], pars=(Deltalnf,refine)
        if freqs[0]=='nyquist_log':
            if freqs[1] is not None:
                Deltatlnf, refine = freqs[1]
            else:
                Deltatlnf, refine = 5e-5, 0
            freqs = pytools.composite_nyquist_log_freq_grid(fLow, fHigh, m1, m2, Deltalnf=Deltatlnf, refine=refine)

    # We do not assume that deltaf is constant
    df = np.diff(freqs)

    # Freq series at base point - force the use of the same gridfreq everywhere
    # gridfreq is a dictionary if hm present, a np array if not
    if modes==[(2,2)]:
        gridfreq = wftdi_base[(2,2)]['freq'].copy()
    else:
        gridfreq = dict([(lm,wftdi_base[lm]['freq'].copy()) for lm in modes])
    waveform_params['gridfreq'] = gridfreq
    # This generates the wf on the input gridfreq, fixed to be the same as base
    # and then interpolates on the large frequency array freqs
    tdifreqseries_base = lisa.GenerateLISATDIFreqseries_SMBH(params_base, freqs, **waveform_params)

    # Get pair of mass/spin params
    # Must have 2 mass, spin params in list_params + list_fixed_params together
    # When varying one param, one needs to specify which 2nd param is fixed
    list_mass_params = [p for p in list_params if p in pytools.list_mass_params]
    list_spin_params = [p for p in list_params if p in pytools.list_spin_params]
    list_fixed_mass_params = [p for p in list_fixed_params if p in pytools.list_mass_params]
    list_fixed_spin_params = [p for p in list_fixed_params if p in pytools.list_spin_params]
    list_pair_mass_params = list_mass_params + list_fixed_mass_params
    list_pair_spin_params = list_spin_params + list_fixed_spin_params
    if (not len(list_mass_params)+len(list_fixed_mass_params)==2) or (not len(list_spin_params)+len(list_fixed_spin_params)==2):
        raise ValueError('list_params + list_fixed_params must have 2 mass and spin params.')
    # The only fully degenerate pair is (q,eta)
    if list_pair_mass_params==['q', 'eta'] or list_pair_mass_params==['eta', 'q']:
        raise ValueError('Mass params pair (q,eta) is fully degenerate.')
    # Remove redundant parameters: keep only relevant pair of mass/spin params
    params_base_pair_massspin = params_base.copy()
    for p in pytools.list_mass_params:
        params_base_pair_massspin.pop(p, None)
    for p in pytools.list_spin_params:
        params_base_pair_massspin.pop(p, None)
    for p in list_pair_mass_params + list_pair_spin_params:
        params_base_pair_massspin[p] = params_base[p]

    # Compute noises
    noise_evaluator = pyLISAnoise.initialize_noise(LISAnoise,
                                        TDI=TDI, TDIrescaled=TDIrescaled,
                                        LISAconst=LISAconst)
    Sn1_vals, Sn2_vals, Sn3_vals = pyLISAnoise.evaluate_noise(
                          LISAnoise, noise_evaluator, freqs,
                          TDI=TDI, TDIrescaled=TDIrescaled, LISAconst=LISAconst)
    Snvals = {}
    Snvals['chan1'] = Sn1_vals
    Snvals['chan2'] = Sn2_vals
    Snvals['chan3'] = Sn3_vals

    # Signals shifted by parameter steps
    steps = dict(steps)
    params_step = {}
    dh = {}
    for p in list_params:
        frac = dict_params_fractional[p]
        params_step = params_base_pair_massspin.copy()
        if frac:
            params_step[p] *= 1 + steps[p]
        else:
            params_step[p] += steps[p]
        params_step = pytools.complete_mass_params(params_step)
        params_step = pytools.complete_spin_params(params_step)

        tdifreqseries_step = lisa.GenerateLISATDIFreqseries_SMBH(params_step, freqs, **waveform_params)

        # Compute signal derivatives dh/dp
        dh[p] = {}
        for chan in channels:
            dh[p][chan] = np.zeros_like(freqs, dtype=complex)
            for lm in modes:
                dh[p][chan] += (tdifreqseries_step[lm][chan] - tdifreqseries_base[lm][chan]) / steps[p]

    # Compute (d1h|d2h)
    F = np.zeros((Npar, Npar), dtype=float)
    for i in range(Npar):
        for j in range(i, Npar):
            Fij = 0.
            pi = list_params[i]
            pj = list_params[j]
            for chan in channels:
                Fij += 4*np.sum(df * np.real(dh[pi][chan] * np.conj(dh[pj][chan]) / Snvals[chan])[:-1])
            F[i,j] = Fij
            #print("Fij:",pi,pj,Fij,Fij*steps[pi]*steps[pj])

    # Symmetrize
    for i in range(Npar):
        for j in range(0, i):
            F[i,j] = F[j,i]

    # Scale back parameters with fractional steps
    scales = np.ones(len(list_params))
    for i,p in enumerate(list_params):
        if dict_params_fractional[p]:
            scales[i] = 1./params_base[p]
    scales_diag = np.diag(scales)
    scales_diag_inv = np.diag(1./scales)

    fisher = np.dot(scales_diag, np.dot(F, scales_diag))
    if prior_invcov is not None:
        F = F + np.dot(scales_diag, np.dot(prior_invcov, scales_diag))
    cov = np.dot(scales_diag_inv, np.dot(np.linalg.inv(F), scales_diag_inv))

    fisher_dict = {}
    fisher_dict['params'] = params.copy()
    fisher_dict['list_params'] = list_params
    fisher_dict['fisher'] = fisher
    fisher_dict['cov'] = cov
    fisher_dict['Lframe'] = Lframe

    return fisher_dict

def fisher_covariance_sobh(params, freqs=None, steps=default_steps, list_params=default_list_fisher_params, prior_invcov=None, Lframe=False, **waveform_params):
    '''
    This is based on fisher_covariance_smbh (as recently enhanced), but adapted for sobh signals
    '''

    waveform_params = copy.deepcopy(waveform_params)

    # Parameters
    Npar = len(list_params)
    m1 = params['m1']
    m2 = params['m2']
    M = m1 + m2
    q = m1 / m2
    params['M'] = M
    params['q'] = q
    Ms = M * pyconstants.MTSUN_SI
    #for SOBH we support Mchirp and eta params
    params['Mchirp'] = pytools.Mchirpofm1m2(m1,m2)
    params['eta']    = pytools.etaofq(q)
    chi1 = params['chi1']
    chi2 = params['chi2']
    fstart = params['fstart']
    dist = params['dist']
    #inc = params['inc']
    #phi = params['phi']
    #lambd = params['lambda']
    #beta = params['beta']
    #psi = params['psi']

    if Lframe:
        raise ValueError('Lframe not implemented (not relevant ?) for SBHBs.')

    LISAnoise = waveform_params.get('LISAnoise', pyLISAnoise.LISAnoiseSciRDv1)
    TDI = waveform_params.get('TDI', 'TDIAET')
    TDIrescaled = waveform_params.get('TDIrescaled', True)
    LISAconst = waveform_params.get('LISAconst', pyresponse.LISAconstProposal)
    minf = waveform_params.get('minf',1e-5)
    maxf = waveform_params.get('maxf',0.5)
    fend=waveform_params.get('fend',None)
    tstart=waveform_params.get('tstart',0.)
    tmin=waveform_params.get('tmin',None)
    tmax=waveform_params.get('tmax',None)
    fref_for_phiref=waveform_params.get('fref_for_phiref',None)
    phiref=waveform_params.get('phiref',0.)

    # Default frequencies (safe and slow):
    # linear spacing, adjust deltaf=1/(2T) with T approximate duration
    if freqs is None:
        freqs = ['linear', None]

    # Determine the frequencies to use for the overlaps, if not given as input
    if not isinstance(freqs, np.ndarray):

        # Determine (2,2) frequency bounds based on frequency and time limits
        fLow = np.fmax(fstart, minf)
        fHigh = maxf
        if fend is not None: fHigh = np.fmin(fHigh, fend)

        # Setting fref for time and phase
        # tref, fref_for_tref are not free, we set tf=0 (tSSB=t0) at fstart
        tref = 0.
        fref_for_tref = fstart
        # Specifying fref_for_phiref, phiref defines the source frame
        # If fref_for_phiref is None (default), use fstart
        if fref_for_phiref is None: fref_for_phiref = fstart
        # No Deltat allowed in principle, since we set the time at fstart
        Deltat = 0.

        # Take into account time cuts if specified
        # TODO: this repeats the initialization of the waveform, not optimal
        # NOTE: the cut in tmax is ignored if tmax > tpeak
        if (tmin is not None) or (tmax is not None):
            fstart_t_acc = 1e-6 # Hardcoded accuracy of f(t) function in s
            mock_gridfreq = np.array([fLow, fHigh])
            mock_wfClass = pyIMRPhenomD.IMRPhenomDh22AmpPhase(mock_gridfreq, m1, m2, chi1, chi2, dist, tref=tref, phiref=phiref, fref_for_tref=fref_for_tref, fref_for_phiref=fref_for_phiref, force_phiref_fref=False, Deltat=Deltat)
            mock_tpeak = mock_wfClass.get_tpeak()
            if tmin is not None:
                tmin_s = tmin * pyconstants.YRSID_SI
                fLow_tmin = mock_wfClass.compute_foft(tmin_s, fLow, fstart_t_acc)
                fLow = np.fmax(fLow, fLow_tmin)
            if tmax is not None:
                tmax_s = tmax * pyconstants.YRSID_SI
                if tmax_s < mock_tpeak:
                    fHigh_tmax = mock_wfClass.compute_foft(tmax_s, fHigh, fstart_t_acc)
                    fHigh = np.fmin(fHigh, fHigh_tmax)

        # Format: freqs = ['log', npt], npt=None for auto setting with T
        if freqs[0]=='log':
            if freqs[1] is not None:
                nptlog = freqs[1]
            else:
                # Estimate length and choose deltaf according to Nyquist
                # NOTE: No safety margin, so set timetomerger_max must carefully
                T = pytools.funcNewtoniantoff(m1, m2, fLow)
                deltaf = 1./(2*T)
                nptlog = np.ceil(np.log(fHigh/fLow) / np.log(1. + deltaf/fLow))
            freqs = pytools.logspace(fLow, fHigh, nptlog)

        # Format: freqs = ['linear', deltaf], deltaf=None for auto setting with T
        if freqs[0]=='linear':
            if freqs[1] is not None:
                deltaf = freqs[1]
            else:
                # Estimate length and choose deltaf according to Nyquist
                # NOTE: No safety margin, so set timetomerger_max must carefully
                T = pytools.funcNewtoniantoff(m1, m2, fLow)
                deltaf = 1./(2*T)
            freqs = np.arange(fLow, fHigh, deltaf)

    # We do not assume that deltaf is constant
    df = np.diff(freqs)

    # Base point - force the use of the same gridfreq everywhere
    params_base = params.copy()
    wftdi_base = lisa.GenerateLISATDI_SOBH(params_base, **waveform_params)
    gridfreq = wftdi_base[(2,2)]['freq'].copy()
    waveform_params['gridfreq'] = gridfreq
    tdifreqseries_base = lisa.GenerateLISATDIFreqseries_SOBH(params_base, freqs, **waveform_params)

    # Signals shifted by parameter steps
    params_step = {}
    tdifreqseries_step = {}
    for p in list_params:
        frac = dict_params_fractional[p]
        params_step = params.copy()
        if frac:
            params_step[p] *= 1 + steps[p]
        else:
            params_step[p] += steps[p]
        if 'M' in list_params:  #Assume using M/q parameters
            M_step = params_step['M']
            q_step = params_step['q']
            params_step['m1'] = M_step * q_step / (1+q_step)
            params_step['m2'] = M_step * 1 / (1+q_step)
        if 'Mchirp' in list_params: #Assume using Mchirp/eta parameters
            Mchirp_step = params_step['Mchirp']
            eta_step = params_step['eta']
            params_step['m1']=pytools.m1ofMchirpeta(Mchirp_step,eta_step)
            params_step['m2']=pytools.m2ofMchirpeta(Mchirp_step,eta_step)
        tdifreqseries_step[p] = lisa.GenerateLISATDIFreqseries_SOBH(params_step, freqs, **waveform_params)

    # Compute noises
    noise_evaluator = pyLISAnoise.initialize_noise(LISAnoise,
                                        TDI=TDI, TDIrescaled=TDIrescaled,
                                        LISAconst=LISAconst)
    Sn1_vals, Sn2_vals, Sn3_vals = pyLISAnoise.evaluate_noise(
                          LISAnoise, noise_evaluator, freqs,
                          TDI=TDI, TDIrescaled=TDIrescaled, LISAconst=LISAconst)

    # Compute derivatives dh
    modes = tdifreqseries_base['modes']  #This should only be (2,2) for SOBH
    channels = ['chan1', 'chan2', 'chan3']
    Snvals = {}
    Snvals['chan1'] = Sn1_vals
    Snvals['chan2'] = Sn2_vals
    Snvals['chan3'] = Sn3_vals
    dh = {}
    for p in list_params:
        dh[p] = {}
        for chan in channels:
            dh[p][chan] = np.zeros_like(freqs, dtype=complex)
            for lm in modes:
                dh[p][chan] += (tdifreqseries_step[p][lm][chan] - tdifreqseries_base[lm][chan]) / steps[p]

    # Compute (d1h|d2h)
    F = np.zeros((Npar, Npar), dtype=float)
    for i in range(Npar):
        for j in range(i, Npar):
            Fij = 0.
            pi = list_params[i]
            pj = list_params[j]
            for chan in channels:
                Fij += 4*np.sum(df * np.real(dh[pi][chan] * np.conj(dh[pj][chan]) / Snvals[chan])[:-1])
            F[i,j] = Fij

    # Symmetrize
    for i in range(Npar):
        for j in range(0, i):
            F[i,j] = F[j,i]

    # Scale back parameters with fractional steps
    scales = np.ones(len(list_params))
    for i,p in enumerate(list_params):
        if dict_params_fractional[p]:
            scales[i] = 1./params[p]
    scales_diag = np.diag(scales)
    scales_diag_inv = np.diag(1./scales)

    fisher = np.dot(scales_diag, np.dot(F, scales_diag))
    if prior_invcov is not None:
        F = F + np.dot(scales_diag, np.dot(prior_invcov, scales_diag))
    cov = np.dot(scales_diag_inv, np.dot(np.linalg.inv(F), scales_diag_inv))

    fisher_dict = {}
    fisher_dict['params'] = params.copy()
    fisher_dict['list_params'] = list_params
    fisher_dict['prior_invcov'] = np.copy(prior_invcov)
    fisher_dict['fisher'] = fisher
    fisher_dict['cov'] = cov
    fisher_dict['Lframe'] = Lframe

    return fisher_dict

################################################################################
# I/O utilities

# Write a fishercov dictionary to hdf5
# NOTE: complications for lists of strings (UTF8/ascii) with h5py ?
def write_h5py_fishercov(h5gr, fishercov):
    params_gr = h5gr.create_group('params')
    pytools.write_h5py_dict(params_gr, fishercov['params'])
    h5gr.create_dataset('list_params', data=np.array(fishercov['list_params'], dtype='S'))
    h5gr.create_dataset('cov', data=fishercov['cov'])
    h5gr.create_dataset('fisher', data=fishercov['fisher'])
    # Need to create a 1-element array for a single value, here boolean
    h5gr.create_dataset('Lframe', data=np.array([fishercov['Lframe']]))
    return
def read_h5py_fishercov(h5gr):
    fishercov = {}
    for k in ['cov', 'fisher']:
        fishercov[k] = h5gr[k][:]
    # Need to read a 1-element array for a single value, here boolean
    fishercov['Lframe'] = h5gr['Lframe'][:][0]
    fishercov['params'] = pytools.read_h5py_dict(h5gr['params'])

    # Need to convert strings from bytestrings to utf-8 -- is this ok ?
    fishercov['list_params'] = [s.decode('utf-8') for s in h5gr['list_params']]
    return fishercov

################################################################################
def lisa_fischer():
    print('''
    Lisa Fischer: How Can I Ease the Pain (1991)
    All alone, on my knees I pray
    For the strength to stay away
    In and out, out and in you go
    I feel your fire
    Then I lose my self control
    How can I ease the pain
    When I know your coming back again
    And how can I ease the pain in my heart
    How can I ease the pain
    Everytime that I let you in
    You take away something deep within
    A fool for love is a fool for pain
    But I refuse to love you again
    How can I ease the pain
    When I know your coming back again
    And how can I ease the pain in my heart
    How can I ease the pain
    If it's not love you've come here for
    Tell me baby why you're here?
    Knock, knock, knockin at my door
    I can't take it, no more
    No more, no more, no more baby
    Give me all or nothing at all
    How can I ease the pain
    When I know your coming back again
    And how can I ease the pain in my heart
    How can I ease the pain
    When I know your coming back again
    And how can I ease the pain in my heart
    I need to know how
    How can I, ease it
    Oo, how can I ease the pain
    I need to know baby
    Oo, how can I ease the pain
    I need to know how to ease it
    How can I ease the pain
    ''')



#
# Copyright (C) 2020 Sylvain Marsat.
#
#


"""
    Python functions for multiemcee simulated inference of MBHBs with LISA.
"""


import os
import copy
import json
import h5py
import argparse
import numpy as np

import lisabeta
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools
import lisabeta.lisa.lisatools as lisatools
import lisabeta.lisa.snrtools as snrtools
import lisabeta.lisa.lisa as lisa
import lisabeta.lisa.lisa_fisher as lisa_fisher
import lisabeta.inference.multiemcee as multiemcee
import lisabeta.inference.inference as inference


################################################################################
# Parse arguments, default params

# Parse arguments
parser = argparse.ArgumentParser(description='Run SMBH parameter inference \
                                              using lisabeta and multiemcee.')
parser.add_argument('input_file', type=str, help='Input json file \
                                                  with all parameters.')
args = parser.parse_args()

# Full list of physical params - infer_params are a subset
# TODO: generalize to other choices of mass and spin parameters
# TODO: at the moment, infer_params can only be a subset of this, not flexible
list_params = [
    "Mchirp",
    "q",
    "chi1",
    "chi2",
    "Deltat",
    "dist",
    "inc",
    "phi",
    "lambda",
    "beta",
    "psi"]

# Default waveform params
waveform_params_default = {
    "minf": 1e-5,
    "maxf": 0.5,
    "t0": 0.0,
    "timetomerger_max": 1.0,
    "fend": None,
    "tmin": None,
    "tmax": None,
    "phiref": 0.0,
    "fref_for_phiref": 0.0,
    "tref": 0.0,
    "fref_for_tref": 0.0,
    "force_phiref_fref": True,
    "toffset": 0.0,
    "modes": None,
    "TDI": "TDIAET",
    "acc": 1e-4,
    "order_fresnel_stencil": 0,
    "approximant": "IMRPhenomD",
    "LISAconst": "Proposal",
    "responseapprox": "full",
    "frozenLISA": False,
    "TDIrescaled": True,
    "LISAnoise": {
    "InstrumentalNoise": "SciRDv1",
    "WDbackground": True,
    "WDduration" : 3.0
    },
    "waveform_params_update_for_template": {}
}

# Default pymultinest params
# TODO: print_info, n_iter_info very rudimentary at the moment
run_params_musthave = ['out_dir', 'out_name']
run_params_default = {
    "sampler": "multiemcee",
    "sample_Lframe": True,
    "multimodal": True,
    "multimodal_pattern": "8modes",
    "p_jump": 0.5,
    "init_scale_cov": 1.,
    "zerolike": False,
    "n_walkers": 100,
    "n_iter": 1000,
    "burn_in": 100,
    "autocor_method": "acor",
    "thin_samples": True,
    "upsample": 1,
    "seed": None,
    "print_info": False,
    "n_iter_info": 10,
    "output": True,
    "output_raw": True
}

# Default parameter range for angular parameters
params_range_default = {
    "inc": [0., np.pi],
    "phi": [-np.pi, np.pi],
    "lambda": [-np.pi, np.pi],
    "beta": [-np.pi/2, np.pi/2],
    "psi": [0., np.pi]
}


################################################################################
# Main

if __name__ == '__main__':

    # Load json file with all parameters
    with open(args.input_file, 'r') as input_file:
        input_params = json.load(input_file)
    # Source params and prior params must be given
    source_params = input_params['source_params']
    source_params = pytools.complete_mass_params(source_params)
    prior_params = input_params['prior_params']
    # Source params can be specified in the Lframe
    source_params_are_Lframe = source_params.get('Lframe', False)
    # Waveform params and run params have predefined default values
    # Need to cast from json to python, list of modes list[list] -> list[tuple]
    waveform_params = waveform_params_default.copy()
    input_params['waveform_params'] = inference.waveform_params_json2py(
                                                input_params['waveform_params'])
    waveform_params.update(input_params['waveform_params'])
    run_params = run_params_default.copy()
    run_params.update(input_params['run_params'])
    zerolike = run_params.pop('zerolike', False)
    sample_Lframe = run_params.pop('sample_Lframe', False)

    # Check run params are meant to be for the multiemcee sampler
    if not (run_params['sampler'] == 'multiemcee'):
        raise ValueError('run_params sampler flag is %s instead of multiemcee.'\
                          % (run_params['sampler']))
    # run params must contain output prefix for files -- no default here
    for musthave in run_params_musthave:
        if not (musthave in run_params):
            raise ValueError('run_params must contain %s.' % (musthave))
    # If n_iter=0, stop at initialization stage; then no burn-in, thinning
    if run_params['n_iter']==0:
        if (not run_params['burn_in']==0) or run_params['thin_samples']:
            raise ValueError('With n_iter=0 (stop at initialization) we cannot \
                                have burn_in or thinning.')
    # Check that the output directory exists
    if (run_params['output'] or run_params['output_raw']):
        if not os.path.isdir(run_params['out_dir']):
            raise ValueError('Output dir %s does not exist.' % run_params['out_dir'])

    # If source params are given in the Lframe, convert to SSB
    if source_params_are_Lframe:
        source_params_SSBframe = lisatools.convert_Lframe_to_SSBframe(
                                    source_params,
                                    t0=waveform_params['t0'],
                                    frozenLISA=waveform_params['frozenLISA'])
        source_params_Lframe = source_params.copy()
    else:
        source_params_SSBframe = source_params.copy()
        source_params_Lframe = lisatools.convert_SSBframe_to_Lframe(
                                    source_params,
                                    t0=waveform_params['t0'],
                                    frozenLISA=waveform_params['frozenLISA'])

    # Use default param ranges to complete input ranges -- check completeness
    prior_type = prior_params['prior_type']
    params_range = prior_params['params_range']
    infer_params = prior_params['infer_params']
    for i, param in enumerate(infer_params):
        if params_range[i] == []:
            params_range[i] = params_range_default[param]
    if [] in params_range:
        raise ValueError('Some parameter ranges are neither given \
                      nor covered by the defaults.')

    # Parameters inferred and parameters fixed
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in list_params if p not in infer_params]
    n_dim = len(infer_params)
    n_fixed = len(fixed_params)
    # Vector x of infer_params for the injection
    # Also store values of fixed params in dict for later use
    if sample_Lframe:
        fixed_params_dict = dict([(p, source_params_Lframe[p]) for p in fixed_params])
        x_inj = np.array([source_params_Lframe[p] for p in infer_params])
    else:
        fixed_params_dict = dict([(p, source_params_SSBframe[p]) for p in fixed_params])
        x_inj = np.array([source_params_SSBframe[p] for p in infer_params])

    # TODO: implement wrapped parameters
    # wrapped_params = [0] * n_dim
    # for i in range(n_dim):
    #     wrapped_params[i] = int(prior_params['wrap_params'][i])
    # run_params['wrapped_params'] = wrapped_params

    # Multimodal jumps
    # TODO: support both 2-mode and 8-mode jumps
    # TODO: be more flexible when some parameters are pinned to their value
    # For now we require all ['inc', 'lambda', 'beta', 'psi'] to be
    # inferred params
    # TODO: improve the back-and-forth between vector and dict
    multimodal = run_params['multimodal']
    multimodal_pattern = run_params['multimodal_pattern']
    dim_extra = 0
    map_extra = None
    proposal_extra = None
    if multimodal:
        if not sample_Lframe:
            raise ValueError('The multimodal option is only supported \
                                together with sample_Lframe.')
        dim_extra = 2

        list_angle_params = ['inc', 'lambda', 'beta', 'psi']
        index_map_angles = {}
        for param in list_angle_params:
            if not param in infer_params:
                raise ValueError('For multimodality, infer_params must contain \
                                        [inc, phi, lambda, beta, psi]')
            index_map_angles[param] = infer_params.index(param)
        # x is array with physical parameters + skymode indices
        # returns x_map, with physical parameters only
        def map_extra(x):
            dim = len(x) - dim_extra
            skymode_index0, skymode_index1 = x[dim:]
            angle_params = {}
            for param in list_angle_params:
                angle_params[param] = x[index_map_angles[param]]
            angle_params_map = inference.map_skymode(angle_params, skymode_index0, skymode_index1)
            x_map = np.copy(x[:dim])
            for param in list_angle_params:
                x_map[index_map_angles[param]] = angle_params_map[param]
            return x_map
        def proposal_extra(x_extra):
            return inference.proposal_skymode(pattern=multimodal_pattern) # in fact independent of skymode indices

    # Prior function (assumes independent multiplicative prior in all params)
    def prior(x):
        p = 1.
        for i in range(n_dim):
            p *= inference.compute_prior(prior_type[i], params_range[i], x[i])
        return p
    # Set up prior bounds for x vector of parameters, dimension (n_dim, 2)
    prior_bounds = np.array(params_range)
    # Check the injection is within the prior bounds
    if not inference.prior_check(x_inj, prior_bounds):
        raise ValueError('Source parameters not within the prior bounds!')

    # Waveform params to be used for template waveforms
    # May differ from the waveform params for injection, given as an update
    waveform_params_template = copy.deepcopy(waveform_params)
    if 'waveform_params_update_for_template' in waveform_params:
        waveform_params_template.update(
                         waveform_params['waveform_params_update_for_template'])

    # Set up likelihood class
    likelihoodClass = lisa.LikelihoodLISASMBH(source_params_Lframe,
                                              **waveform_params)

    # Define likelihood function, wrapped for use in multiemcee
    # TODO: add m1>m2 to prior range checks
    def lnlikelihood(x):
        if not inference.prior_check(x, prior_bounds):
            return -1e99
        template_params = {}
        for i in range(n_dim):
            template_params[infer_params[i]] = x[i]
        for p in fixed_params:
            template_params[p] = fixed_params_dict[p]
        template_params = pytools.complete_mass_params(template_params)
        template_params = pytools.complete_spin_params(template_params)
        template_params['Lframe'] = sample_Lframe
        lnL = likelihoodClass.lnL(template_params, **waveform_params_template)
        return lnL

    # Read run_params
    init_scale_cov = run_params['init_scale_cov']
    p_extra = run_params['p_jump']
    n_walkers = run_params['n_walkers']
    n_iter = run_params['n_iter']
    burn_in = run_params['burn_in']
    autocor_method = run_params['autocor_method']
    upsample = run_params['upsample']
    seed = run_params['seed']
    print_info = run_params['print_info']
    n_iter_info = run_params['n_iter_info']

    if print_info:
        print('Source parameters SSBframe:')
        print(source_params_SSBframe)
        print('Source parameters Lframe:')
        print(source_params_Lframe)

    # Compute Fisher matrix
    # NOTE: might be slow if timetomerger_max is set at generic value like 1yr
    # NOTE: numerical problems in Fisher when timetomerger_max is too large
    # For Fisher only, we set timetomerger_max based on SNR=1 threshold
    waveform_params_22 = waveform_params.copy()
    waveform_params_22['modes'] = [(2,2)]
    t_SNR1_s = snrtools.lisa_tofSNR(1., source_params_SSBframe, **waveform_params_22)
    t_SNR1_yr = t_SNR1_s / pyconstants.YRSID_SI
    # Ensure at least 1d of signal for fisher regardless of SNR threshold
    t_threshold_yr = np.fmax(t_SNR1_yr, 1./365.25)
    waveform_params_fisher = waveform_params.copy()
    waveform_params_fisher['timetomerger_max'] = t_threshold_yr
    # TODO: improve handling of different choices of infer_params
    if print_info:
        print('Computing Fisher matrix...')
        print('Threshold max(t(SNR=1), 1d) used for Fisher: %g yr' % (t_threshold_yr))
    fishercov = lisa_fisher.fisher_covariance_smbh(source_params_SSBframe,
                        steps=lisa_fisher.default_steps, freqs=['log', None],
                        list_params=infer_params, list_fixed_params=fixed_params,
                        Lframe=sample_Lframe,
                        **waveform_params_fisher)

    if print_info:
        print('Fisher covariance matrix:')
        print(fishercov['cov'])

    # Initialize points drawing from the Fisher matrix
    x_ini = np.zeros((n_walkers, n_dim + dim_extra))
    if sample_Lframe:
        mean = np.array([source_params_Lframe[p] for p in infer_params])
    else:
        mean = np.array([source_params_SSBframe[p] for p in infer_params])
    # We allow for the possibility of scaling the covariance for initialization
    cov = fishercov['cov'] * init_scale_cov
    try:
        for i in range(n_walkers):
            x_ini[i,:n_dim] = inference.draw_multivariate_normal_constrained(
                                          mean, cov, params_range, infer_params)
    except:
        # Regularizing Fisher with a prior inverse covariance in spins
        print('Fisher initialization failed, adding regularizing prior invcov at sigma=0.5 for spins.')
        # By default, add a Gaussian prior with sigma=0.5 in both spins
        fisher_prior_invcov = np.zeros((n_dim,n_dim), dtype=float)
        if 'chi1' in infer_params:
            ichi1 = infer_params.index('chi1')
            fisher_prior_invcov[ichi1,ichi1] = 1./0.5**2
        if 'chi2' in infer_params:
            ichi2 = infer_params.index('chi2')
            fisher_prior_invcov[ichi2,ichi2] = 1./0.5**2
        fishercov_reg = lisa_fisher.fisher_covariance_smbh(source_params_SSBframe,
                            steps=lisa_fisher.default_steps, freqs=['log', None],
                            list_params=infer_params, list_fixed_params=fixed_params,
                            Lframe=sample_Lframe, prior_invcov=fisher_prior_invcov,
                            **waveform_params_fisher)
        print('Fisher covariance matrix, regularized in spins:')
        print(fishercov_reg['cov'])
        cov_reg = fishercov_reg['cov'] * init_scale_cov
        try:
            for k in range(n_temps):
                for i in range(n_walkers):
                    x_ini[k,i,:] = inference.draw_multivariate_normal_constrained(
                                          mean, cov_reg, params_range, infer_params)
        except:
            print('Fisher initialization failed with regularization, initializing from prior.')
            for j in range(n_dim):
                x_ini[:,j] = inference.draw_prior(prior_type[j], prior_bounds[j], n_walkers)
    # Initial points all start on the main mode (1,0)
    if multimodal:
        x_ini[:,n_dim] = 1
        x_ini[:,n_dim+1] = 0
    # Enforce range of phase parameters in case Fisher uncertainties are large
    # TODO: we should have proper rejection based on the prior ranges
    if 'inc' in infer_params:
        iinc = infer_params.index('inc')
        x_ini[:,iinc] = pytools.modpi(x_ini[:,iinc])
    if 'phi' in infer_params:
        iphi = infer_params.index('phi')
        x_ini[:,iphi] = pytools.mod2pi(x_ini[:,iphi])
    if 'lambda' in infer_params:
        ilambda = infer_params.index('lambda')
        x_ini[:,ilambda] = pytools.mod2pi(x_ini[:,ilambda])
    if 'beta' in infer_params:
        ibeta = infer_params.index('beta')
        x_ini[:,ibeta] = pytools.modpi2(x_ini[:,ibeta])
    if 'psi' in infer_params:
        ipsi = infer_params.index('psi')
        x_ini[:,ipsi] = pytools.mod2pi(x_ini[:,ipsi])

    # A la grace de dieu
    if print_info:
        print('Running multiemcee...')
    chain, lnlikevals, lnpostvals, acceptvals = multiemcee.multiemcee_run(
        lnlikelihood, prior, n_walkers, n_iter, x_ini,
        prior_bounds=prior_bounds, dim_extra=dim_extra, p_extra=p_extra,
        map_extra=map_extra, proposal_extra=proposal_extra, seed=seed,
        print_info=print_info, n_iter_info=n_iter_info)

    # TODO: output meta information, acceptance rate and likelihood speed

    # Post-processing: unfold if multimodal was used
    if print_info:
        print('Post-processing...')
    if multimodal:
        # For niter=0, we unfold/output the initial samples
        chain_unfold_shape = (n_walkers, max(n_iter, 1), n_dim)
        chain_unfold = np.zeros(chain_unfold_shape, dtype=float)
        for k in range(n_walkers):
            chain_unfold[k] = multiemcee.unfold_chain(chain[k],
                                                      dim_extra, map_extra)
    else:
        chain_unfold = chain

    # Post-processing: burn-in, autocor. length and thinning, merge walkers
    # Also complete with params that were fixed to their values
    # Only if niter>0; if niter=0 we stop at initialization stage for testing
    thin_len = 1
    if n_iter>0:
        if burn_in>=n_iter:
            print('WARNING: burn_in is larger than n_iter, ignoring burn_in.')
            burn_in = 0
        if run_params['thin_samples']:
            autocor_len = {}
            for i,param in enumerate(infer_params):
                # Note: autocorr_new computes a mean across walkers
                autocor_len[param] = multiemcee.get_autocor(chain_unfold[:,burn_in:,i], method=autocor_method)
                if print_info:
                    print('Autocorrelation length for %s: %g' % (param,autocor_len[param]))
            # The thinning length is the worse autocor length for all params
            thin_len = int(np.ceil(np.max([autocor_len[param] for param in infer_params])))
            thin_len = int(np.ceil(thin_len / float(upsample)))
        if print_info:
            print('Autocorrelation thinning: %d' % (thin_len,))
    # Processed output: thinned and merged chain (completed with fixed params)
    if n_iter>0:
        n_out = n_walkers * ((n_iter-burn_in-1) // thin_len + 1)
    else:
        n_out = n_walkers
    chain_processed = np.zeros((n_out, n_dim + n_fixed), dtype=float)
    # Here ravel with order='F' flattens by reading along columns first
    for i in range(n_dim):
        i_out = list_params.index(infer_params[i])
        chain_processed[:,i_out] = np.ravel(chain_unfold[:,burn_in::thin_len,i],
                                            order='F')
    for i in range(n_fixed):
        i_out = list_params.index(fixed_params[i])
        if sample_Lframe:
            chain_processed[:,i_out] = source_params_Lframe[fixed_params[i]]
        else:
            chain_processed[:,i_out] = source_params_SSBframe[fixed_params[i]]
    # Thin and merge likelihood, posterior, acceptance vals
    lnlikevals_processed = np.ravel(lnlikevals[:,burn_in::thin_len], order='F')
    lnpostvals_processed = np.ravel(lnpostvals[:,burn_in::thin_len], order='F')
    acceptvals_processed = np.ravel(acceptvals[:,burn_in::thin_len], order='F')

    # Output
    basename = run_params['out_dir'] + run_params['out_name']
    # Main output: processed, thinned chains and Fisher matrix
    # TODO: hierarchy, create a level (group ?) samples
    if run_params['output']:
        with h5py.File(basename + '.h5', 'w') as f:
            # Source params
            source_params_Lframe_gr = f.create_group('source_params_Lframe')
            pytools.write_h5py_dict(source_params_Lframe_gr, source_params_Lframe)
            source_params_SSBframe_gr = f.create_group('source_params_SSBframe')
            pytools.write_h5py_dict(source_params_SSBframe_gr, source_params_SSBframe)
            for i,param in enumerate(list_params):
                f.create_dataset(param, data=chain_processed[:,i])
            f.create_dataset('lnlike', data=lnlikevals_processed)
            f.create_dataset('lnpost', data=lnpostvals_processed)
            f.create_dataset('accept', data=acceptvals_processed)
            fishercov_gr = f.create_group('fishercov')
            lisa_fisher.write_h5py_fishercov(fishercov_gr, fishercov)
    # Extra output: full chains, walker-separated, for testing and diagnostics
    if run_params['output_raw']:
        with h5py.File(basename + '_raw.h5', 'w') as f:
            for i,param in enumerate(infer_params):
                f.create_dataset(param, data=chain[:,:,i])
            f.create_dataset('lnlike', data=lnlikevals)
            f.create_dataset('lnpost', data=lnpostvals)
            f.create_dataset('accept', data=acceptvals)



from abc import ABC, abstractmethod
import numpy as np
import lisabeta
import lisabeta.pyconstants as pyconstants
from scipy import interpolate
import os

# C = constants.Nature
# CLIGHT = C.VELOCITYOFLIGHT_CONSTANT_VACUUM
# year = C.SIDEREALYEAR_J2000DAY*24*60*60
CLIGHT = pyconstants.C_SI
year =  pyconstants.YRSID_SI


def simple_snr(f, h, i=None, years=1.0, noise_model='SciRDv1'):
    if not i:
        h0 = h * np.sqrt(16.0/5.0)    # rms average over inclinations
    else:
        h0 = h * np.sqrt((1 + np.cos(i)**2)**2 + (2.*np.cos(i))**2)
    sens = get_noise_model(noise_model,f, wd=years).sensitivity()
    snr = h0 * np.sqrt(years * 365.25*24*3600) / np.sqrt(sens)
    return snr


def get_noise_model(model, frq=None, **kwargs):
    """Return the noise instance corresponding to model.

    model can be: "Proposal", "SciRDv1", "SciRDdeg1", "MRDv1",
    "mldc", "newdrs", "LCESAcall"
    """
    if model in ["Proposal", "SciRDv1", "MRDv1"]:
        NoiseClass = globals()["AnalyticNoise"]
        return NoiseClass(frq, model, **kwargs)
    elif model=="mldc":
        NoiseClass = globals()[model.upper()+"Noise"]
        return NoiseClass(frq, **kwargs)
    elif model=="SciRDdeg1":
        NoiseClass = globals()["SciRDdeg1Noise"]
        return NoiseClass(frq, **kwargs)
    elif model=="SciRDdeg2":
        NoiseClass = globals()["SciRDdeg2Noise"]
        return NoiseClass(frq, **kwargs)
    elif model=="newdrs":
        NoiseClass = globals()["NewRDSNoise"]
        return NoiseClass(frq, **kwargs)
    elif os.path.exists(model):
        NoiseClass = globals()["NumericNoise"]
        return NumericNoise.from_file(model)#NoiseClass(frq, model)
    else:
        NoiseClass = globals()[model+"Noise"]
        return NoiseClass(frq, **kwargs)

class Noise():

    def __init__(self, frq, wd=0):
        self.freq = frq
        self.wd = wd

    def set_wdconfusion(self, duration):
        day = 24.*60*60
        if ((duration!=0 and duration < day/year) or (duration > 10.0)):
            raise NotImplementedError
        self.wd = duration

    @property
    def arm_length(self):
        return  2.5e9 # LISA's arm meters

    def WDconfusion(self, freq=None, option="X"):
        """ option can be X or AE
        """
        if freq is None:
            freq = self.freq
        duration = self.wd
        if duration==0:
            return 0
        lisaLT = self.arm_length/CLIGHT
        x = 2.0 * np.pi * lisaLT * freq
        t = 4.0 * x**2 * np.sin(x)**2
        Sg_sens = GalNoise(freq, duration*year).galshape()
        SgX = t*Sg_sens
        if option=="A" or option=="E":
            return 1.5*SgX
        elif option=="XY":
            return -0.5*SgX
        else:
            return SgX

    def to_file(self, filename):
        """ Save noise PSD to file
        """
        TDIn = np.rec.fromarrays([self.freq,
                                  self.psd(option="X"),
                                  self.psd(option="XY")], names=["freq", "X","XY"])
        np.save(filename, TDIn)

    def _XY2AET(self, X, XY, option='A'):
        if option in ["A", "E"]:
            return X-XY # = E
        elif option in ["T"]:
            return X+2*XY




class NumericNoise(Noise):
    """ PSD from file.
    """
    def __init__(self, psdarray):
        self._psd = np.empty_like(psdarray)
        self.freq = psdarray["freq"]
        for k in ["X", "XY"]:
            self._psd[k] = psdarray[k]
        self.wd = 0

    @staticmethod
    def from_file(filename):
        psd = np.load(filename)
        assert 'freq' in psd.dtype.names
        assert 'X' in psd.dtype.names
        assert 'XY' in psd.dtype.names
        N = NumericNoise(psd)
        return N

    def psd(self, freq=None, option='X'):
        if option in ['A', 'E', 'T']:
            XX = self.psd(freq, option="X")
            XY = self.psd(freq, option="XY")
            return self._XY2AET(XX, XY, option)

        if freq is not None:
            if np.isscalar(freq):
                freq = np.array([freq])
            S = np.interp(freq, self.freq, self._psd[option])
        if self.wd:
            S += self.WDconfusion(freq=freq, option=option)
        return S

class AnalyticNoise(Noise):
    """Analytic approximation of the two components of LISA noise:
    acceleration noise and optical metrology system (OMS) noise
    """

    def __init__(self, frq, model="SciRDv1", wd=0):

        Noise.__init__(self, frq, wd=wd)
        Sloc = (1.7e-12)**2 # m^2/Hz
        Ssci = (8.9e-12)**2 # m^2/Hz
        Soth = (2.e-12)**2  # m^2/Hz

        self.DSoms_d = {'Proposal':(10.e-12)**2,
                       'SciRDv1':(15.e-12)**2,
                       'MRDv1':(10.e-12)**2}  # m^2/Hz
        self.DSa_a = {'Proposal':(3.e-15)**2,
                     'SciRDv1':(3.e-15)**2,
                     'MRDv1':(2.4e-15)**2} # m^2/sec^4/Hz

        # Acceleration noise
        Sa_a = self.DSa_a[model] *(1.0 +(0.4e-3/frq)**2)*(1.0+(frq/8e-3)**4) # in acceleration
        self.Sa_d = Sa_a*(2.*np.pi*frq)**(-4.) # in displacement

        # In relative frequency unit
        Sa_nu = self.Sa_d*(2.0*np.pi*frq/CLIGHT)**2
        self.Spm =  Sa_nu

        # Optical Metrology System
        self.Soms_d = self.DSoms_d[model] * (1. + (2.e-3/frq)**4) # In displacement
        Soms_nu = self.Soms_d*(2.0*np.pi*frq/CLIGHT)**2 # In relative frequency unit
        self.Sop =  Soms_nu
        self.freq = frq
        self.model = model

    def relative_freq(self):
        return self.Spm, self.Sop

    def displacement(self):
        return self.Sa_d, self.Soms_d

    def sensitivity(self):#, includewd=0):
        ALL_m = np.sqrt( 4.0*self.Sa_d + self.Sop)
        AvResp = np.sqrt(5)         ## Average the antenna response
        Proj = 2./np.sqrt(3)     ## Projection effect

        ## Approximative transfert function
        f0 = 1.0/(2.0*(self.arm_length/CLIGHT))
        a = 0.41
        T  = np.sqrt(1+(self.freq/(a*f0))**2)
        Sens = (AvResp * Proj * T * ALL_m/self.arm_length)**2

        if (self.wd):
            day = 86400.0
            if ((self.wd < day/year) or (self.wd > 10.0)):
                raise NotImplementedError
            Sgal = GalNoise(self.freq, self.wd*year).galshape()
            Sens += Sgal
        return Sens

    def psd(self, freq=None, option="X"):#, includewd=0):
        """ option can be X, X2, XY, A, E, T
        """
        if option in ['A', 'E', 'T']:
            XX = self.psd(freq, option="X")
            XY = self.psd(freq, option="XY")
            return self._XY2AET(XX, XY, option)

        if self.wd and option in ["X2", "T"]:
            raise NotImplementedError

        if freq is not None:
            self.__init__(freq, model=self.model, wd=self.wd)

        lisaLT = self.arm_length/CLIGHT
        x = 2.0 * np.pi * lisaLT * self.freq
        if option=="X":
            S = 16.0 * np.sin(x)**2 * (2.0 * (1.0 + np.cos(x)**2) * self.Spm + self.Sop)
        elif option=="X2":
            S = 64.0 * np.sin(x)**2 * np.sin(2*x)**2 * self.Sop # TODO Check the acc. noise term
            S += 256.0 * (3 + np.cos(2*x)) * np.cos(x)**2 * np.sin(x)**4 * self.Spm
        elif option=="XY":
            S = -4.0 * np.sin(2*x) * np.sin(x) * (self.Sop + 4.0*self.Spm)
        elif option=="A" or option=="E":
            S = 8.0 * np.sin(x)**2 * (2.0 * self.Spm * (3.0 + 2.0*np.cos(x) +\
                                                        np.cos(2*x)) +\
                                      self.Sop * (2.0 + np.cos(x)))
        elif option=="T":
            S = 16.0 * self.Sop * (1.0 - np.cos(x)) * np.sin(x)**2 +\
                128.0 * self.Spm * np.sin(x)**2 * np.sin(0.5*x)**4
        else:
            print("PSD option should be in [X, X2, XY, A, E, T] (%s)"%option)
            return None
        if self.wd:
            S += self.WDconfusion(freq=freq, option=option)
        return S



class SciRDdeg1Noise(AnalyticNoise):
    """
    """
    def __init__(self, frq, model="SciRDv1", wd=0):

        AnalyticNoise.__init__(self, frq, model, wd=wd)
        Sa_a = self.DSa_a['SciRDv1'] * (1.0+(0.4e-3/frq)**2) * \
               (1.0+(frq/8e-3)**4) * (1.0+(0.1e-3/frq)**2)
        self.Sa_d = Sa_a*(2.*np.pi*frq)**(-4.)
        Sa_nu = self.Sa_d*(2.0*np.pi*frq/CLIGHT)**2
        self.Spm =  Sa_nu


class SciRDdeg2Noise(AnalyticNoise):
    """
    """
    def __init__(self, frq, model="SciRDv1", wd=0):

        AnalyticNoise.__init__(self, frq, model, wd=wd)
        Sa_a = self.DSa_a['SciRDv1'] * (1.0+(0.4e-3/frq)**2) * \
               (1.0+(frq/8e-3)**4) * (1.0+(0.1e-3/frq)**4)
        self.Sa_d = Sa_a*(2.*np.pi*frq)**(-4.)
        Sa_nu = self.Sa_d*(2.0*np.pi*frq/CLIGHT)**2
        self.Spm =  Sa_nu

class MLDCNoise(AnalyticNoise):
    """
    """
    def __init__(self, f, wd=0):
        Noise.__init__(self, frq, wd=wd)
        self.Spm = 2.5e-48 * (1.0 + (f/1.0e-4)**-2) * f**(-2)
        defaultL = 16.6782
        lisaLT = self.arm_length/CLIGHT
        self.Sop = 1.8e-37 * (lisaLT/defaultL)**2 * f**2
        self.freq = f

    def displacement(self):
        raise NotImplementedError("")

class LCESAcallNoise(AnalyticNoise):
    """
    """
    def __init__(self, frq, wd=0):
        AnalyticNoise.__init__(self, frq, wd=wd)

        # Acceleration noise
        Sa_a = self.DSa_a['Proposal'] *(1.0 +(0.4e-3/frq)**2+(frq/9.3e-3)**4) # in acceleration
        self.Sa_d = Sa_a*(2.*np.pi*frq)**(-4.) # in displacement
        Sa_nu = self.Sa_d*(2.0*np.pi*frq/CLIGHT)**2 # in relative frequency unit
        self.Spm =  Sa_nu

        # Optical Metrology System
        self.Soms_d = self.DSoms_d['Proposal'] * (1. + (2.e-3/frq)**4) # in displacement
        Soms_nu = self.Soms_d*(2.0*np.pi*frq/CLIGHT)**2 # in relative frequency unit
        self.Sop =  Soms_nu
        self.freq = frq

class NewDRSNoise(AnalyticNoise):
    """
    """
    def __init__(self, f, wd=0):
        AnalyticNoise.__init__(self, frq, wd=wd)
        self.Spm = 6.00314e-48 * f**(-2) # 4.6e-15 m/s^2/sqrt(Hz)
        defaultL = 16.6782
        defaultD = 0.4
        defaultP = 1.0

        lisaD = 0.3  # TODO check it
        lisaP = 2.0  # TODO check it
        lisaLT = self.arm_length/CLIGHT
        Sops = 6.15e-38 * (lisaLT/defaultL)**2 * (defaultD/lisaD)**4 *\
               (defaultP/lisaP) # 11.83 pm/sqrt(Hz)
        Sopo = 2.81e-38 # 8 pm/sqrt(Hz)
        self.Sop = (Sops + Sopo) * f**2
        self.freq = f

    def displacement(self):
        raise NotImplementedError("")


class GalNoise():
    """
    Here is an empirical fit for the Galactic confusion as a function of observation time
    We have used SNR>5.0 and SNR>7 as removal criteria.
    We use default to be 6 kinks, and used combined SNR. We also provide fits for 4 links
    @Tobs duration of observation (in seconds) should be between 3 months and 10 years
    """

    def __init__(self, freq, Tobs, SNR=7.0, links=6):

        self.freq = freq
        Tobs = Tobs/year

        if (SNR != 5.0 and SNR != 7.0):
            print ('We accept SNR to be 5.0 or 7.0', 'given', SNR)
            raise NotImplementedError

        self.Ampl, self.alpha, self.fr2, af1, bf1, afk, bfk = [ 1.28265531e-44, 1.62966700e+00,  4.81078093e-04, -2.23499956e-01,\
                                                                -2.70408439e+00, -3.60976122e-01, -2.37822436e+00]
        if (SNR == 5.0):
            self.Ampl, self.alpha, self.fr2, af1, bf1, afk, bfk =  [ 1.28265531e-44,  1.57926625e+00,  3.84466946e-04, -2.32198526e-01,\
                                                                    -2.77607250e+00, -2.78383832e-01, -2.51448127e+00]
        if (links == 4):
            self.Ampl, self.alpha, self.fr2, af1, bf1, afk, bfk =  [ 1.28265531e-44, 1.62966700e+00,  4.81078093e-04, -2.62516578e-01, \
                                                                    -2.60291132e+00, -4.20352978e-01, -2.23847341e+00]
            if (SNR == 5.0):
                self.Ampl, self.alpha, self.fr2, af1, bf1, afk, bfk =  [ 1.28265531e-44,  1.57926625e+00,  3.84466946e-04, -2.39762426e-01, -2.70133615e+00,\
                                                                        -3.49472046e-01, -2.37368814e+00]
        Tmin = 0.25
        Tmax = 10.0
        if (Tobs<Tmin or Tobs>Tmax):
            print ('Galaxy fit is valid between 3 months and 10 years, we do not extrapolate', Tobs, ' not in', Tmin, Tmax)
            raise NotImplementedError("")

        self.fr1 = 10.**(af1*np.log10(Tobs) + bf1)
        self.fknee = 10.**(afk*np.log10(Tobs) + bfk)


    def galshape(self):
        res = self.Ampl*np.exp(-(self.freq/self.fr1)**self.alpha) *\
              (self.freq**(-7./3.))*0.5*(1.0 + np.tanh(-(self.freq-self.fknee)/self.fr2))
        return res



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython definitions for LISA FD response.
"""

import numpy as np
cimport numpy as np
from enum import Enum

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *


cdef extern from "LISAgeometry.h":

    enum TDItag:
        TDIAET = 0
        TDIXYZ = 1
        TDI2AET = 2
        TDI2XYZ = 3

    enum ResponseApproxtag:
        full         = 0
        lowfL        = 1
        lowf         = 2
        ignoreRdelay = 3

    ctypedef struct LISAconstellation:
        double OrbitOmega;
        double OrbitPhi0;
        double Orbitt0;
        double OrbitR;
        double OrbitL;

cdef extern from "LISAresponse.h":

    int EvalLISAFDresponseTDI3Chan (
        real_vector** phaseRdelay,
        complex_vector** transfer1,
        complex_vector** transfer2,
        complex_vector** transfer3,
        real_vector* freq,
        real_vector* tf,
        const double t0,
        const int l,
        const int m,
        const double inc,
        const double phi,
        const double lambd,
        const double beta,
        const double psi,
        const TDItag tditag,
        const LISAconstellation *LISAconst,
        const ResponseApproxtag responseapprox,
        const int tagfrozenLISA,
        const int rescaled
    );



#
# Copyright (C) 2020 Sylvain Marsat.
#
#


"""
    Python tools for Bayesian inference.
"""


import time
import numpy as np
import copy
import h5py

import lisabeta
import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pytools as pytools

# Make this a soft dependency -- will only be needed if the user asks for this
# tool in computing autocorrelations
try:
    import acor
except ImportError:
    pass


################################################################################
# Wrap parameters

# Params that are wrapped -- by default we wrap psi mod pi
# phi is wrapped mod 2pi by default for the general HM case
default_wrap_dict = {
    'phi': [-np.pi, np.pi],
    'lambda': [-np.pi, np.pi],
    'ra': [-np.pi, np.pi],
    'psi': [0., np.pi],
    }
def func_default_wrap(list_params):
    list_wrap_moduli = []
    for p in list_params:
        list_wrap_moduli += [default_wrap_dict.get(p, None)]
    return list_wrap_moduli

################################################################################
# Priors

def prior_check(x, prior_bounds):
    if prior_bounds is None:
        return True
    else:
        return np.all([(prior_bounds[i][0]<=x[i] and x[i]<=prior_bounds[i][1]) for i in range(len(x))])

# Computing simple priors
# Assumes paramrange[0] < paramrange[1]
# TODO: add support for non-trivial mass priors (flat in m1/m2 -> Mchirp/eta)
# TODO: add automatic ranges for dimensionful parameters like masses, distance
# TODO: for now the extra options are clunky (also unnormalized)
# ever-growing if/elif -  we should map once and for all to a function
def compute_prior(priortype, paramrange, x):
    if priortype=='uniform':
        return 1. / (paramrange[1] - paramrange[0])
    elif priortype=='sin': # NOTE: cos decreasing on [0,pi]
        return np.sin(x) / (np.cos(paramrange[0]) - np.cos(paramrange[1]))
    elif priortype=='cos': # NOTE: sin increasing on [-pi/2,pi/2]
        return np.cos(x) / (np.sin(paramrange[1]) - np.sin(paramrange[0]))
    elif priortype=='quadratic':
        return x*x / 3. / (paramrange[1]**3 - paramrange[0]**3)
    elif priortype=='Mchirp_Mchirpqchipchim_flatlnMinvqchi1chi2':
        return 1e6/x # Unnormalized Mchirpref/Mchirp, arbitrary Mchirpref
    elif priortype=='q_Mchirpqchipchim_flatlnMinvqchi1chi2':
        return (1+x)**2/x**3 # Unnormalized (1+q)**2/q**3
    else:
        raise ValueError('priortype not recognized.')

# Set priors as pure functions
# Avoids going through if/elif at evaluation, still multiplicative only
def def_prior_func(priortype, paramrange):
    if priortype=='uniform':
        norm = 1 / (paramrange[1] - paramrange[0])
        return lambda x: 1. * norm
    elif priortype=='sin': # NOTE: cos decreasing on [0,pi]
        norm = 1 / (np.cos(paramrange[0]) - np.cos(paramrange[1]))
        return lambda x: np.sin(x) * norm
    elif priortype=='cos': # NOTE: sin increasing on [-pi/2,pi/2]
        norm  = 1 / (np.sin(paramrange[1]) - np.sin(paramrange[0]))
        return lambda x: np.cos(x) * norm
    elif priortype=='quadratic':
        norm = 1 / 3. / (paramrange[1]**3 - paramrange[0]**3)
        return lambda x: x*x * norm
    elif priortype=='Mchirp_Mchirpqchipchim_flatlnMinvqchi1chi2':
        return lambda x: 1e6/x # Unnormalized Mchirpref/Mchirp, arbitrary Mchirpref
    elif priortype=='q_Mchirpqchipchim_flatlnMinvqchi1chi2':
        return lambda x: (1+x)**2/x**3 # Unnormalized (1+q)**2/q**3
    else:
        raise ValueError('priortype not recognized.')

# Draw from simple priors
def draw_prior(priortype, paramrange, n):
    u = np.random.uniform(size=n)
    if priortype=='uniform':
        return paramrange[0] + u * (paramrange[1] - paramrange[0])
    elif priortype=='sin': # NOTE: cos decreasing on [0,pi]
        return np.arccos(np.cos(paramrange[1]) + u * (np.cos(paramrange[0]) - np.cos(paramrange[1])))
    elif priortype=='cos': # NOTE: sin increasing on [-pi/2,pi/2]
        return np.arcsin(np.sin(paramrange[0]) + u * (np.sin(paramrange[1]) - np.sin(paramrange[0])))
    elif priortype=='quadratic':
        return np.power(paramrange[0]**3 + u * (paramrange[1]**3 - paramrange[0]**3), 1./3)
    else:
        raise ValueError('priortype not recognized for drawing.')

################################################################################
# Autocorrelation
# see https://dfm.io/posts/autocorr/
# see https://github.com/dfm/emcee/issues/209

def next_pow_two(n):
    i = 1
    while i < n:
        i = i << 1
    return i

def autocorr_func_1d(x, norm=True):
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2*n)
    acf = np.fft.ifft(f * np.conjugate(f))[:len(x)].real
    acf /= 4*n

    # Optionally normalize
    if norm:
        acf /= acf[0]

    return acf

# Automated windowing procedure following Sokal (1989)
def auto_window(taus, c):
    m = np.arange(len(taus)) < c * taus
    if np.any(m):
        return np.argmin(m)
    return len(taus) - 1

# Following the suggestion from Goodman & Weare (2010)
def autocorr_gw2010(y, c=5.0):
    f = autocorr_func_1d(np.mean(y, axis=0))
    taus = 2.0*np.cumsum(f)-1.0
    window = auto_window(taus, c)
    return taus[window]

def autocorr_new(y, c=5.0):
    f = np.zeros(y.shape[1])
    for yy in y:
        f += autocorr_func_1d(yy)
    f /= len(y)
    taus = 2.0*np.cumsum(f)-1.0
    window = auto_window(taus, c)
    return taus[window]

# Wrapper, allowing two choices, for data given as an array of chains
def get_autocor(data, method='autocor_new'):
    if method=='autocor_new':
        return autocorr_new(data)
    elif method=='acor':
        return np.mean([acor.acor(data[k])[0] for k in range(len(data))])
    else:
        raise ValueError('Autocorrelation method %s not recognized.' % (method))

################################################################################
# Constrained Gaussian draws
# Recursively draw from Gaussian until within the prior bounds

# Using infer_params to identify angular params, first map to 2pi or pi-range
# Necessary as Fisher errors for e.g. phi, psi can be very large
# TODO: a bit clumsy, but this is used at initialization essentially
def enforce_range_angles(x, infer_params):
    # Handle both 1d and 2d arrays, y is always 2d
    if x.ndim==1:
        y = np.array([x])
    else:
        y = x.copy()
    if 'inc' in infer_params:
        iinc = infer_params.index('inc')
        y[:,iinc] = pytools.modpi(y[:,iinc])
    if 'phi' in infer_params:
        iphi = infer_params.index('phi')
        y[:,iphi] = pytools.mod2pi(y[:,iphi])
    if 'lambda' in infer_params:
        ilambda = infer_params.index('lambda')
        y[:,ilambda] = pytools.mod2pi(y[:,ilambda])
    if 'beta' in infer_params:
        ibeta = infer_params.index('beta')
        y[:,ibeta] = pytools.modpi2(y[:,ibeta])
    if 'psi' in infer_params:
        ipsi = infer_params.index('psi')
        y[:,ipsi] = pytools.modpi(y[:,ipsi])
    if x.ndim==1:
        return y[0]
    else:
        return y

def draw_multivariate_normal_constrained(mean, cov, prior_bounds, infer_params):
    x = np.random.multivariate_normal(mean, cov)
    x = enforce_range_angles(x, infer_params)
    if prior_check(x, prior_bounds):
        return x
    else:
        return draw_multivariate_normal_constrained(mean, cov, prior_bounds, infer_params)

################################################################################
# Interface for json

def waveform_params_json2py(waveform_params):
    waveform_params_py = copy.deepcopy(waveform_params)
    if ('modes' in waveform_params.keys()) and isinstance(waveform_params_py['modes'], list):
        waveform_params_py['modes'] = [tuple(lm) for lm in waveform_params['modes']]
    if 'waveform_params_update_for_template' in waveform_params.keys():
        if ('modes' in waveform_params['waveform_params_update_for_template'].keys()) and isinstance(waveform_params['waveform_params_update_for_template']['modes'], list):
            waveform_params_py['waveform_params_update_for_template']['modes'] = [tuple(lm) for lm in waveform_params['waveform_params_update_for_template']['modes']]
    if 'gridfreq' in waveform_params.keys():
        waveform_params_py['gridfreq'] = np.array(waveform_params['gridfreq'])
    return waveform_params_py

################################################################################
# Functions for multimodal jumps, see Marsat&al arXiv:2003.00357

def map_skymode(angle_params, skymode_index0, skymode_index1):
    inc = angle_params['inc']
    lambdaL = angle_params['lambda']
    betaL = angle_params['beta']
    psiL = angle_params['psi']
    jump_params = {}
    jump_params['inc'] = pytools.modpi(np.pi/2 - (skymode_index0 *(np.pi/2 - inc)))
    jump_params['lambda'] = pytools.mod2pi(lambdaL + skymode_index1 * np.pi/2)
    jump_params['beta'] = skymode_index0 * betaL
    jump_params['psi'] = pytools.modpi(np.pi/2 - (skymode_index0 *(np.pi/2 - psiL)) + skymode_index1 * np.pi/2)
    return jump_params

# Used e.g. for the simple22 map: transform sky and indexpsi
# In the inverse map, indexpsi=[0,1] decides of the branch psi \in [0,pi/2] or [pi/2,pi]
def map_skymode_simple22map(angle_params, skymode_index0, skymode_index1, skymode_index2):
    lambdaL = angle_params['lambda']
    sbetaL = angle_params['sbeta']
    indexpsi = angle_params['indexpsi']
    jump_params = {}
    jump_params['lambda'] = pytools.mod2pi(lambdaL + skymode_index1 * np.pi/2)
    jump_params['sbeta'] = skymode_index0 * sbetaL
    jump_params['indexpsi'] = skymode_index2
    return jump_params

def proposal_skymode(size=None, pattern='8modes', random_state=np.random):
    if pattern=='8modes':
        return np.array([2 * (np.random.binomial(1, 0.5, size=size) - 1./2), np.random.randint(0, 4, size=size)]).T
    elif pattern=='2modes':
        return np.array([2 * (np.random.binomial(1, 0.5, size=size) - 1./2), np.zeros(size, dtype=int)]).T
    else:
        raise ValueError('pattern %s not recognized.' % (pattern))

# Used e.g. for the simple22 map: transform sky and indexpsi
# In the inverse map, indexpsi=[0,1] decides of the branch psi \in [0,pi/2] or [pi/2,pi]
def proposal_skymode_simple22map(size=None, pattern='8modes', random_state=np.random):
    if pattern=='8modes':
        return np.array([2 * (np.random.binomial(1, 0.5, size=size) - 1./2), np.random.randint(0, 4, size=size), np.random.randint(0, 2, size=size)]).T
    elif pattern=='2modes':
        return np.array([2 * (np.random.binomial(1, 0.5, size=size) - 1./2), np.zeros(size, dtype=int), np.random.randint(0, 2, size=size)]).T
    else:
        raise ValueError('pattern %s not recognized.' % (pattern))

################################################################################
# Functions for a parameter map adapted to the simple likelihood with 22 mode
# see Marsat&al arXiv:2003.00357

# def func_extrparams(params, injparams):
#     params_extr = {}
#     for p in ['inc', 'phi', 'lambda', 'beta', 'psi']:
#         params_extr[p] = params[p]
#     params_extr['d'] = params['dist'] / injparams['dist']
#     params_extr['phi'] = pytools.modpi(params_extr['phi'])
#     return params_extr

# def func_restoreparams(params, injparams):
#     params_restore = {}
#     for p in ['inc', 'phi', 'lambda', 'beta', 'psi']:
#         params_restore[p] = params[p]
#     params_restore['dist'] = params['d'] * injparams['dist']
#     params_restore['phi'] = pytools.modpi(params['phi'])
#     for p in pytools.list_mass_params:
#         params_restore[p] = injparams[p]
#     for p in pytools.list_spin_params:
#         params_restore[p] = injparams[p]
#     params_restore['Deltat'] = injparams['Deltat']
#     params_restore['Lframe'] = injparams['Lframe']
#     return params_restore

def map_params_simple_response_22(params, injdist):
    dist = params['dist']
    inc = params['inc']
    phi = params['phi']
    lambda_a = params['lambda'] - np.pi/6
    beta = params['beta']
    psi = params['psi']

    if (np.any(dist<=0.) or np.any(inc<0.) or np.any(inc>np.pi) or np.any(beta<-np.pi/2) or np.any(beta>np.pi/2)):
        raise ValueError('params are outside physical range.')

    d = dist / injdist
    tiota = np.tan(inc/2)
    ttheta = np.tan(1./2 * (np.pi/2 - beta))
    rho = 1./(4*d) * np.sqrt(5/np.pi) * 1./(1+tiota**2)**2 * 1./(1+ttheta**2)**2

    sigma_plus = rho*np.exp(2*1j*phi) * (ttheta**4 * np.exp(-2*1j*psi) + tiota**4 * np.exp(2*1j*psi)) * np.exp(-2*1j*lambda_a)
    sigma_minus = rho*np.exp(2*1j*phi) * (np.exp(-2*1j*psi) + ttheta**4 * tiota**4 * np.exp(2*1j*psi)) * np.exp(2*1j*lambda_a)

    if isinstance(psi, np.ndarray):
        indexpsi = np.zeros(psi.shape, dtype=int)
        indexpsi[pytools.modpi(psi) > np.pi/2] = 1
    else:
        if pytools.modpi(psi) <= np.pi/2:
            indexpsi = 0
        else:
            indexpsi = 1

    params_map = {}
    params_map['Aplus'] = np.abs(sigma_plus)
    params_map['Aminus'] = np.abs(sigma_minus)
    params_map['Phiplus'] = np.angle(sigma_plus)
    params_map['Phiminus'] = np.angle(sigma_minus)
    params_map['lambda'] = lambda_a + np.pi/6
    params_map['sbeta'] = np.sin(beta)
    params_map['indexpsi'] = indexpsi

    return params_map

def physical_check_simple_response_22(params_map):

    Aplus = params_map['Aplus']
    Aminus = params_map['Aminus']
    sbeta = params_map['sbeta']
    indexpsi = params_map['indexpsi']

    return (not (np.any(Aplus<=0.) or np.any(Aminus<=0.) or np.any(sbeta<-1.) or np.any(sbeta>1.) or not np.all((indexpsi==0) | (indexpsi==1))))

def invmap_params_simple_response_22(params_map, injdist):

    Aplus = params_map['Aplus']
    Aminus = params_map['Aminus']
    Phiplus = params_map['Phiplus']
    Phiminus = params_map['Phiminus']
    lambd = params_map['lambda']
    sbeta = params_map['sbeta']
    indexpsi = params_map['indexpsi']

    if (np.any(Aplus<=0.) or np.any(Aminus<=0.) or np.any(sbeta<-1.) or np.any(sbeta>1.) or not np.all((indexpsi==0) | (indexpsi==1))):
        raise ValueError('params_map are outside physical range.')

    lambda_a = lambd - np.pi/6
    sigma_plus = Aplus * np.exp(1j*Phiplus)
    sigma_minus = Aminus * np.exp(1j*Phiminus)
    rtilde = sigma_plus / sigma_minus * np.exp(4*1j*lambda_a)

    ctheta = sbeta
    a = ((ctheta-1) / (ctheta+1))**2
    bz = (a - rtilde) / (a*rtilde - 1)
    b = np.abs(bz)
    tiota2 = np.sqrt(b)
    ciota = (1 - tiota2) / (1 + tiota2)
    fourpsi = np.angle(bz)
    # Ambiguity in psi: psi0 in [0, pi/2], psi1=psi0+pi/2 in [pi/2, pi]
    # We use indexpsi=[0,1] to represent this degeneracy
    psi0 = pytools.mod_interval(1./4*fourpsi, interval=[0, np.pi/2])
    psi = psi0 + indexpsi*np.pi/2

    # Get a,b,z
    a = ((ctheta-1) / (ctheta+1))**2
    b = ((ciota-1) / (ciota+1))**2
    z = np.exp(1j*4*psi)

    # No ambiguity in phi, since for 22-mode only we can restrict it mod pi
    phi = 1./2 * np.angle(sigma_minus / (np.exp(-2*1j*psi) * (1+ a*b*z) * np.exp(1j*2*lambda_a)))
    phi = pytools.modpi(phi)

    # Amplitude, unambiguous
    rho = np.abs(sigma_minus / (np.exp(-2*1j*psi) * (1+ a*b*z) * np.exp(1j*2*lambda_a)))
    ttheta2 = np.sqrt(a)
    d = 1/(4*rho) * np.sqrt(5/np.pi) * 1./((1+ttheta2)**2 * (1+tiota2)**2)
    beta = np.arcsin(sbeta)
    inc = np.arccos(ciota)

    params = {}
    params['dist'] = d * injdist
    params['inc'] = inc
    params['phi'] = phi
    params['lambda'] = lambd
    params['beta'] = beta
    params['psi'] = psi

    return params

def jacobian_map_simple_response_22(params, params_map, injdist):

    d = params['dist'] / injdist
    inc = params['inc']
    beta = params['beta']

    Aplus = params_map['Aplus']
    Aminus = params_map['Aminus']

    ttheta2 = np.tan(1./2 * (np.pi/2-beta))**2
    tiota2 = np.tan(1./2 * inc)**2

    return (5/np.pi)**2 * 1./(32*d**3) * 1./(Aplus*Aminus) * (1+ttheta2**2)**2 * (1-ttheta2)**2 * tiota2**3 / (1+ttheta2)**6 / (1+tiota2)**6

def complete_params_degen_response_22(params, injdist):

    params_map = map_params_simple_response_22(params, injdist)

    params_complete = copy.deepcopy(params)
    for p in ['Aplus', 'Aminus', 'Phiplus', 'Phiminus', 'indexpsi']:
        params_complete[p] = params_map[p]
    params_complete['Phis'] = pytools.mod_interval(1./2 * (params_complete['Phiplus'] + params_complete['Phiminus']), [-np.pi/2, np.pi/2])
    params_complete['Phia'] = pytools.mod_interval(1./2 * (params_complete['Phiplus'] - params_complete['Phiminus']), [-np.pi/2, np.pi/2])

    return params_complete



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Standalone EOBNRv2HMROM inspiral-merger-ringdown GW waveform model
    for binary black hole coalescences.
"""

from __future__ import print_function

import numpy as np
cimport numpy as np
from math import cos
import copy

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *


cdef extern from "struct.h":

    ctypedef struct ListAmpPhaseFDMode:
        AmpPhaseFDMode*              hlm;
        int                          l;
        int                          m;
        ListAmpPhaseFDMode*          next;

    ListAmpPhaseFDMode* ListAmpPhaseFDMode_GetMode(
    	   ListAmpPhaseFDMode* list,  # List structure to get this mode from
    	   int l,                     # Mode number l
    	   int m);                    # Mode number m

    ListAmpPhaseFDMode* ListAmpPhaseFDMode_Destroy(
    	   ListAmpPhaseFDMode* list); # List structure to destroy

cdef extern from "EOBNRv2HMROM.h":
    int SimEOBNRv2HMROMExtTF2(
        ListAmpPhaseFDMode** listhlm, # Output: list of modes, FD amp/phase
        int nbmode,                   # Number of modes to generate
        double Mf_match,              # Match freq TF2/EOBNRv2HMROM (geom units)
        double minf,                  # Minimum frequency required (Hz)
        int tagexthm,                 # Tag for extension of the higher modes
        double deltatRef,             # Time shift (peak of h22 at deltatRef)
        double phiRef,                # Phase at reference frequency
        double fRef,                  # Ref frequency (Hz); 0 for max Mf of ROM
        double m1SI,                  # Mass of companion 1 (kg)
        double m2SI,                  # Mass of companion 2 (kg)
        double distance,              # Distance of source (m)
        int setphiRefatfRef);         # Flag: setting FD phase to phiRef at fRef

cdef class EOBNRv2HMROM:
    """ Generate EOBNRv2HMROM IMR frequency-domain waveform with higher modes.
    """

    cdef ListAmpPhaseFDMode* listhlm  # pointer to waveform structure

    cdef m1_SI                # Mass of companion 1 (kg)
    cdef m2_SI                # Mass of companion 2 (kg)
    cdef dist_SI              # Distance of source (m)
    cdef f_min                # Minimum frequency required (Hz)
    cdef deltatRef            # Time shift (peak of h22 at deltatRef)
    cdef phiRef               # Reference phase at fRef (rad)
    cdef fRef                 # Reference frequency (Hz)
    cdef public object modes  # List of modes (l,m)

    cdef hlm                  # Dictionary of modes

    #cdef modes_default
    modes_ROM = [(2,2), (2,1), (3,3), (4,4), (5,5)]

    def __init__(self,
                 m1_SI, m2_SI, dist_SI, f_min, deltatRef, phiRef, fRef,
                 modes=modes_ROM):
        """Constructor
        Arguments:
          m1_SI                 # Mass of companion 1 (kg)
          m2_SI                 # Mass of companion 2 (kg)
          dist_SI               # Distance of source (m)
          f_min                 # Minimum frequency required (Hz)
          deltatRef             # Time shift (peak of h22 at deltatRef)
          phiRef                # Reference phase at fRef (rad)
          fRef                  # Reference frequency (Hz)
        Keyword arguments:
          modes                 # List of modes to return (all are generated)
        """
        # arguments are checked in the C waveform generator
        self.m1_SI = m1_SI
        self.m2_SI = m2_SI
        self.f_min = f_min
        self.deltatRef = deltatRef
        self.phiRef = phiRef
        self.fRef = fRef
        self.dist_SI = dist_SI
        self.modes = modes

        self.listhlm = NULL

        # We generate all modes, then select user-specified modes for output
        cdef int nbmode = 5
        cdef double Mf_match = 0.
        cdef int tagexthm = 1
        cdef setphiRefatfRef = 1
        ret = SimEOBNRv2HMROMExtTF2(
            &self.listhlm,
            nbmode,
            Mf_match,
            f_min,
            tagexthm,
            deltatRef,
            phiRef,
            fRef,
            m1_SI,
            m2_SI,
            dist_SI,
            setphiRefatfRef
        );
        if ret != 0:
            raise ValueError("Call to SimEOBNRv2HMROMExtTF2() failed.")

        # Read the modes in a dictionary
        # Direct copy of C double array to numpy via a MemoryView
        cdef AmpPhaseFDMode* Chlm = NULL
        self.hlm = {}
        cdef double[::1] view_freq_amp
        cdef double[::1] view_amp_real
        cdef double[::1] view_amp_imag
        cdef double[::1] view_freq_phase
        cdef double[::1] view_phase
        for lm in self.modes:
            (l,m) = (lm[0], lm[1])
            self.hlm[lm] = {}
            if not lm in self.modes_ROM:
                raise ValueError('Mode not allowed: (%d, %d)' % (l, m))
            Chlm = ListAmpPhaseFDMode_GetMode(self.listhlm, l, m).hlm
            view_freq_amp = \
                        <(double)[:Chlm.freq_amp.size]> Chlm.freq_amp.data
            view_amp_real = \
                        <(double)[:Chlm.amp_real.size]> Chlm.amp_real.data
            view_amp_imag = \
                        <(double)[:Chlm.amp_imag.size]> Chlm.amp_imag.data
            view_freq_phase = \
                        <(double)[:Chlm.freq_phase.size]> Chlm.freq_phase.data
            view_phase = \
                        <(double)[:Chlm.phase.size]> Chlm.phase.data
            self.hlm[lm]['freq_amp'] = np.asarray(view_freq_amp)
            self.hlm[lm]['amp_real'] = np.asarray(view_amp_real)
            self.hlm[lm]['amp_imag'] = np.asarray(view_amp_imag)
            self.hlm[lm]['freq_phase'] = np.asarray(view_freq_phase)
            self.hlm[lm]['phase'] = np.asarray(view_phase)

    def __dealloc__(self):
        """Destructor
        """
        if self.listhlm != NULL:
            ListAmpPhaseFDMode_Destroy(self.listhlm)

    # NOTE: to return a copy of the modes, copy.deepcopy is noticeably slower
    def get_waveform(self):
        hlm = {}
        for lm in self.modes:
            (l,m) = (lm[0], lm[1])
            hlm[lm] = {}
            hlm[lm]['freq_amp'] = np.copy(self.hlm[lm]['freq_amp'])
            hlm[lm]['amp_real'] = np.copy(self.hlm[lm]['amp_real'])
            hlm[lm]['amp_imag'] = np.copy(self.hlm[lm]['amp_imag'])
            hlm[lm]['freq_phase'] = np.copy(self.hlm[lm]['freq_phase'])
            hlm[lm]['phase'] = np.copy(self.hlm[lm]['phase'])
        return hlm
